from __future__ import annotations

import numpy as np
import pandas as pd
import xarray as xr
from scipy import interpolate

from sasktran2.geodetic import WGS84
from sasktran2.geometry import Geometry1D, GeometryType, InterpolationMethod
from sasktran2.solar import SolarGeometryHandlerBase

from .base import ViewingGeometryContainer
from .wrappers import TangentAltitudeSolar, ViewingGeometry


class LimbVertical(ViewingGeometry, ViewingGeometryContainer):
    def __init__(
        self,
        solar_handler: SolarGeometryHandlerBase,
        tangent_altitudes: np.array,
        tangent_latitude: float | np.ndarray,
        tangent_longitude: float | np.ndarray,
        time: pd.Timestamp | np.ndarray,
        observer_altitude: float | np.ndarray,
        observer_latitude: float | np.ndarray,
        observer_longitude: float | np.ndarray,
        reference_altitude: float = 25000,
    ):
        """
        A Limb Vertical image is a geometry container where the observer is looking through the limb of the atmosphere, at a set
        of tangent altitudes.

        We recommend constructing this class through one of the various `from_` class methods.  No validation checking
        is done through this constructor.

        Parameters
        ----------
        solar_handler : sk.solar.SolarGeometryHandlerBase
            Solar geometry handler to use for calculating solar angles.
        tangent_altitudes : np.array
            Altitudes of the tangent points.
        tangent_latitude : float | np.ndarray
            Latitude of the tangent points. Can be a scalar or same size as `tangent_altitudes`.
        tangent_longitude : float | np.ndarray
            Longitude of the tangent points. Can be a scalar or same size as `tangent_altitudes`.
        time : pd.Timestamp | np.ndarray
            Time of the observation. Can be a scalar or same size as `tangent_altitudes`.
        observer_altitude : float | np.ndarray
            Altitude of the observer. Can be a scalar or same size as `tangent_altitudes`.
        observer_latitude : float | np.ndarray
            Latitude of the observer. Can be a scalar or same size as `tangent_altitudes`.
        observer_longitude : float | np.ndarray
            Longitude of the observer. Can be a scalar or same size as `tangent_altitudes`.
        reference_altitude : float, optional
            The tangent altitude where reference parameters such as earth radius
            are defined at, by default 25000.
        """
        self._tangent_altitudes = np.array(tangent_altitudes)

        # Determine the expected length
        n = len(self._tangent_altitudes)

        # Helper function to convert scalars to arrays
        def to_array(param):
            if np.isscalar(param) | isinstance(param, pd.Timestamp):
                return np.full(n, param)
            param = np.array(param)
            if len(param) != n:
                msg = "Parameter length mismatch."
                raise ValueError(msg)
            return param

        self._tangent_latitude = to_array(tangent_latitude)
        self._tangent_longitude = to_array(tangent_longitude)
        self._time = to_array(time)
        self._observer_altitude = to_array(observer_altitude)
        self._observer_latitude = to_array(observer_latitude)
        self._observer_longitude = to_array(observer_longitude)

        self._reference_altitude = reference_altitude

        ViewingGeometry.__init__(self)

        tangent_geo = WGS84()
        observer_geo = WGS84()

        viewing_zenith = np.zeros(n)
        viewing_azimuth = np.zeros(n)

        self._cos_sza = np.zeros(n)
        self._earth_radius = np.zeros(n)
        self._solar_azimuth = np.zeros(n)
        self._observer_azimuth = np.zeros(n)

        for i, (alt, lat, lon, t, obs_alt, obs_lat, obs_lon) in enumerate(
            zip(
                tangent_altitudes,
                self._tangent_latitude,
                self._tangent_longitude,
                self._time,
                self._observer_altitude,
                self._observer_latitude,
                self._observer_longitude,
                strict=False,
            )
        ):
            tangent_geo.from_lat_lon_alt(lat, lon, alt)
            observer_geo.from_lat_lon_alt(obs_lat, obs_lon, obs_alt)

            solar_zenith, solar_azimuth = solar_handler.target_solar_angles(
                lat, lon, alt, t
            )

            # Need to calculate the relative azimuth
            lv = tangent_geo.location - observer_geo.location
            lv /= np.linalg.norm(lv)

            # Need angle in the (north, west) plane measured clockwise
            observer_azimuth = -np.arctan2(
                np.dot(lv, tangent_geo.local_west), -np.dot(lv, tangent_geo.local_south)
            )

            # Now observer_azimuth is pointing away from the observer, and solar azimuth is pointing towards the sun
            # If they are equal this is forward scatter as expected,

            self.add_ray(
                TangentAltitudeSolar(
                    tangent_altitude_m=alt,
                    relative_azimuth=np.deg2rad(solar_azimuth - observer_azimuth),
                    observer_altitude_m=obs_alt,
                    cos_sza=np.cos(np.deg2rad(solar_zenith)),
                )
            )

            self._observer_azimuth[i] = observer_azimuth
            self._solar_azimuth[i] = solar_azimuth

            # Assign the viewing zenith and azimuthal angles
            viewing_zenith[i] = np.rad2deg(np.arccos(np.dot(lv, observer_geo.local_up)))

            viewing_azimuth[i] = -np.arctan2(
                np.dot(lv, observer_geo.local_west),
                -np.dot(lv, observer_geo.local_south),
            )

            # Store parameters
            self._cos_sza[i] = np.cos(np.deg2rad(solar_zenith))
            self._earth_radius[i] = np.linalg.norm(
                tangent_geo.location - alt * tangent_geo.local_up
            )

        geometry_ds = xr.Dataset(
            {
                "tangent_altitude": (["los"], tangent_altitudes),
                "tangent_longitude": (
                    ["los"],
                    np.ones_like(tangent_altitudes) * tangent_longitude,
                ),
                "tangent_latitude": (
                    ["los"],
                    np.ones_like(tangent_altitudes) * tangent_latitude,
                ),
                "time": (["los"], self._time.astype(np.datetime64)),
                "observer_altitude": (["los"], self._observer_altitude),
                "observer_latitude": (["los"], self._observer_latitude),
                "observer_longitude": (["los"], self._observer_longitude),
                "tangent_cos_sza": (["los"], self._cos_sza),
                "tangent_solar_azimuth": (["los"], self._solar_azimuth),
                "tangent_observer_azimuth": (["los"], self._observer_azimuth),
                "viewing_zenith": (["los"], viewing_zenith),
                "viewing_azimuth": (["los"], viewing_azimuth),
            },
        )

        ViewingGeometryContainer.__init__(self, geometry_ds)

    def recommended_cos_sza(self) -> float:
        """
        Returns the cosine of the solar zenith angle at the reference altitude.

        Returns
        -------
        float
        """
        return interpolate.interp1d(self._tangent_altitudes, self._cos_sza)(
            self._reference_altitude
        )

    def recommended_earth_radius(self) -> float:
        """
        Returns the earth radius at the reference altitude.

        Returns
        -------
        float
        """
        return interpolate.interp1d(self._tangent_altitudes, self._earth_radius)(
            self._reference_altitude
        )

    def model_geometry(self, altitude_grid_m: np.ndarray) -> Geometry1D:
        return Geometry1D(
            self.recommended_cos_sza(),
            0.0,
            self.recommended_earth_radius(),
            altitude_grid_m,
            InterpolationMethod.LinearInterpolation,
            GeometryType.Spherical,
        )

    @classmethod
    def from_tangent_parameters(
        cls,
        solar_handler: SolarGeometryHandlerBase,
        tangent_altitudes: np.ndarray,
        tangent_latitude: float,
        tangent_longitude: float,
        time: pd.Timestamp,
        observer_altitude: float,
        viewing_azimuth: float,
        reference_altitude: float = 25000,
        forced_constant_tangent: bool = False,
    ):
        """
        Initialize a LimbVertical object from a set of tangent parameters assuming a vertical image is taken. From
        a single observer location.

        Parameters
        ----------
        solar_handler : sk.solar.SolarGeometryHandlerBase
            Solar geometry handler to use for calculating solar angles.
        tangent_altitudes : np.ndarray
            The altitudes of the tangent points.
        tangent_latitude : float
            Latitude of the tangent point at reference_altitude
        tangent_longitude : float
            Longitude of the tangent piont at reference_altitude
        time : pd.Timestamp
            Time of the observation
        observer_altitude : float
            Altitude of the observer
        viewing_azimuth : float
            Azimuth of the viewing direction in degrees clockwise from north at the tangent point.
        reference_altitude : float, optional
            Altitude the tangent point parameters are defined at, by default 25000
        forced_constant_tangent : bool, optional
            If true, then variation in latitude/longitude is not taken into account as a function of
            tangent altitude, by default False
        """
        tangent_geo = WGS84()

        tangent_geo.from_lat_lon_alt(
            tangent_latitude, tangent_longitude, reference_altitude
        )

        lv = -tangent_geo.local_south * np.cos(
            np.deg2rad(viewing_azimuth)
        ) - tangent_geo.local_west * np.sin(np.deg2rad(viewing_azimuth))

        # Now we have to find the observer location
        observer_geo = WGS84()
        observer_geo.from_xyz(
            observer_geo.altitude_intercepts(
                observer_altitude, tangent_geo.location, lv
            )[0]
        )

        # Then we can calculate all of the actual tangent latitudes and longitudes
        actual_tangent_latitude = np.zeros(len(tangent_altitudes))
        actual_tangent_longitude = np.zeros(len(tangent_altitudes))

        if forced_constant_tangent:
            actual_tangent_latitude = tangent_latitude
            actual_tangent_longitude = tangent_longitude
        else:
            for i, alt in enumerate(tangent_altitudes):
                tangent_geo.from_tangent_altitude(alt, observer_geo.location, lv)
                actual_tangent_latitude[i] = tangent_geo.latitude
                actual_tangent_longitude[i] = tangent_geo.longitude

        return cls(
            solar_handler,
            tangent_altitudes,
            actual_tangent_latitude,
            actual_tangent_longitude,
            time,
            observer_altitude,
            observer_geo.latitude,
            observer_geo.longitude,
            reference_altitude,
        )



# ruff: noqa: F401
from __future__ import annotations

from .distribution import LogNormalDistribution, integrate_mie
from .refractive import RefractiveIndex
from .wrappers import LinearizedMie, MieOutput



from __future__ import annotations

from sasktran2._core_rust import (
    EmissionSource,
    InputValidationMode,
    MultipleScatterSource,
    OccultationSource,
    PyConfig,
    SingleScatterSource,
    StokesBasis,
    ThreadingLib,
    ThreadingModel,
)


class Config:
    _config: PyConfig

    def __init__(self):
        """
        Object which stores all of the configuration settings for the radiative transfer calculation.
        """
        self._config = PyConfig()

    def _into_rust_object(self):
        """
        Returns the rust object which contains all of the configuration settings for the radiative transfer calculation.
        """
        return self._config

    @property
    def num_threads(self) -> int:
        """
        Controls the number of threads used in the calculation.  For maximum performance it is
        recommended to set this to the number of physical cores on your machine.  Defaults to
        1
        """
        return self._config.num_threads

    @num_threads.setter
    def num_threads(self, value: int):
        """
        Controls the number of threads used in the calculation.  For maximum performance it is
        recommended to set this to the number of physical cores on your machine.  Defaults to
        1
        """
        self._config.num_threads = value

    @property
    def threading_model(self) -> ThreadingModel:
        """
        Sets the multi-threading mode to use in the calculation.

        `sasktran2.ThreadingModel.Wavelength` (Default)
            Calculation is multi-threaded over the wavelength (batch) dimension only. This method
            works very well when this dimension is large, but may increase memory usage. It also
            is not very effective for a small number of wavelengths.

        `sasktran2.ThreadingModel.Source`
            Calculation is multi-threaded individually by each source function for each wavelength.
            This method is recommended when memory is a concern, or when the number of wavelengths
            is small.
        """
        return self._config.threading_model

    @threading_model.setter
    def threading_model(self, value: ThreadingModel):
        """
        Sets the multi-threading mode to use in the calculation.

        `sasktran2.ThreadingModel.Wavelength` (Default)
            Calculation is multi-threaded over the wavelength (batch) dimension only. This method
            works very well when this dimension is large, but may increase memory usage. It also
            is not very effective for a small number of wavelengths.

        `sasktran2.ThreadingModel.Source`
            Calculation is multi-threaded individually by each source function for each wavelength.
            This method is recommended when memory is a concern, or when the number of wavelengths
            is small.
        """
        self._config.threading_model = value

    @property
    def input_validation_mode(self) -> InputValidationMode:
        """
        Sets the input validation mode to use in the calculation.

        `sasktran2.InputValidationMode.Strict` (Default)
            All input validation checks are performed. This is the recommended mode for most users.

        `sasktran2.InputValidationMode.Standard`
            Only the most important input validation checks are performed. This mode is recommended
            for advanced users who are confident in the input data.

        `sasktran2.InputValidationMode.Disabled`
            No input validation checks are performed. This mode is recommended for advanced users who
            are confident in the input data and want to maximize performance.
        """
        return self._config.input_validation_mode

    @input_validation_mode.setter
    def input_validation_mode(self, value: InputValidationMode):
        """
        Sets the input validation mode to use in the calculation.

        `sasktran2.InputValidationMode.Strict` (Default)
            All input validation checks are performed. This is the recommended mode for most users.

        `sasktran2.InputValidationMode.Standard`
            Only the most important input validation checks are performed. This mode is recommended
            for advanced users who are confident in the input data.

        `sasktran2.InputValidationMode.Disabled`
            No input validation checks are performed. This mode is recommended for advanced users who
            are confident in the input data and want to maximize performance.
        """
        self._config.input_validation_mode = value

    @property
    def num_stokes(self) -> int:
        """
        Sets the number of Stokes parameters used in the calculation. 1 is equivalent to the scalar approximation.
        Currently the only supported values are 1, and 3.
        """
        return self._config.num_stokes

    @num_stokes.setter
    def num_stokes(self, value: int):
        """
        Sets the number of Stokes parameters used in the calculation. 1 is equivalent to the scalar approximation.
        Currently the only supported values are 1, and 3.
        """
        self._config.num_stokes = value

    @property
    def single_scatter_source(self) -> SingleScatterSource:
        """
        Sets which (if any) single scatter source is to be used inside the calculation.

        `sasktran2.SingleScatterSource.Exact` (Default)
            A single scatter source where exact ray tracing is performed at each quadrature
            point along the observer lines of sight towards the sun

        `sasktran2.SingleScatterSource.Table`
            A single scatter source where a pre-computed table is used to calculate solar
            transmission to quadrature points along the line of sight.

        `sasktran2.SingleScatterSource.DiscreteOrdinates`
            Lets the discrete ordinates source function calculate the single scatter source. Only
            has an effect if the geometry mode is set to PlaneParallel or PseudoSpherical, and
            if the DiscreteOrdinates source function is also used for multiple scatter.

        `sasktran2.SingleScatterSource.NoSource`
            Disables the single scatter source
        """
        return self._config.single_scatter_source

    @single_scatter_source.setter
    def single_scatter_source(self, value: SingleScatterSource):
        """
        Sets which (if any) single scatter source is to be used inside the calculation.

        `sasktran2.SingleScatterSource.Exact` (Default)
            A single scatter source where exact ray tracing is performed at each quadrature
            point along the observer lines of sight towards the sun

        `sasktran2.SingleScatterSource.Table`
            A single scatter source where a pre-computed table is used to calculate solar
            transmission to quadrature points along the line of sight.

        `sasktran2.SingleScatterSource.DiscreteOrdinates`
            Lets the discrete ordinates source function calculate the single scatter source. Only
            has an effect if the geometry mode is set to PlaneParallel or PseudoSpherical, and
            if the DiscreteOrdinates source function is also used for multiple scatter.

        `sasktran2.SingleScatterSource.NoSource`
            Disables the single scatter source
        """
        self._config.single_scatter_source = value

    @property
    def occultation_source(self) -> OccultationSource:
        """
        Sets which (if any) occultation source is to be used inside the calculation.

        `sasktran2.OccultationSource.NoSource` (Default)
            No occultation source included

        `sasktran2.OccultationSource.Standard`
            A constant source of 1 is placed at the end of every individual line of sight.
        """
        return self._config.occultation_source

    @occultation_source.setter
    def occultation_source(self, value: OccultationSource):
        """
        Sets which (if any) occultation source is to be used inside the calculation.

        `sasktran2.OccultationSource.NoSource` (Default)
            No occultation source included

        `sasktran2.OccultationSource.Standard`
            A constant source of 1 is placed at the end of every individual line of sight.
        """
        self._config.occultation_source = value

    @property
    def emission_source(self) -> EmissionSource:
        """
        Sets which (if any) emission source is to be used inside the calculation.

        `sasktran2.EmissionSource.NoSource` (Default)
            No emission source included

        `sasktran2.EmissionSource.Standard`
            An emission source defined on the atmosphere grid is enabled.
        """
        return self._config.emission_source

    @emission_source.setter
    def emission_source(self, value: EmissionSource):
        """
        Sets which (if any) emission source is to be used inside the calculation.

        `sasktran2.EmissionSource.NoSource` (Default)
            No emission source included

        `sasktran2.EmissionSource.Standard`
            An emission source defined on the atmosphere grid is enabled.
        """
        self._config.emission_source = value

    @property
    def multiple_scatter_source(self) -> MultipleScatterSource:
        """
        Sets which (if any) multiple scatter source is to be used inside the calculation.

        `sasktran2.MultipleScatterSource.NoSource` (Default)
            Multiple scattering is disabled

        `sasktran2.MultipleScatterSource.DiscreteOrdinates`
            The discrete ordinates technique is used to estimate the multiple scatter signal

        `sasktran2.MultipleScatterSource.SuccessiveOrders`
            The successive orders of scattering method is used to estimate the multiple scatter
            signal

        """
        return self._config.multiple_scatter_source

    @multiple_scatter_source.setter
    def multiple_scatter_source(self, value: MultipleScatterSource):
        """
        Sets which (if any) multiple scatter source is to be used inside the calculation.

        `sasktran2.MultipleScatterSource.NoSource` (Default)
            Multiple scattering is disabled

        `sasktran2.MultipleScatterSource.DiscreteOrdinates`
            The discrete ordinates technique is used to estimate the multiple scatter signal

        `sasktran2.MultipleScatterSource.SuccessiveOrders`
            The successive orders of scattering method is used to estimate the multiple scatter
            signal

        """
        self._config.multiple_scatter_source = value

    @property
    def stokes_basis(self) -> StokesBasis:
        """ """
        return self._config.stokes_basis

    @stokes_basis.setter
    def stokes_basis(self, value: StokesBasis):
        """ """
        self._config.stokes_basis = value

    @property
    def delta_m_scaling(self) -> bool:
        """
        Controls whether the delta-M scaling is applied to the calculation.  Defaults to False.
        """
        return self._config.delta_m_scaling

    @delta_m_scaling.setter
    def delta_m_scaling(self, value: bool):
        """
        Controls whether the delta-M scaling is applied to the calculation.  Defaults to False.
        """
        self._config.delta_m_scaling = value

    @property
    def los_refraction(self) -> bool:
        """
        Controls whether or not refraction is enabled for the observer line of sight rays. Requires
        the refractive index to be set in the Geometry object for refraction to work.  Defaults to False.
        """
        return self._config.los_refraction

    @los_refraction.setter
    def los_refraction(self, value: bool):
        """
        Controls whether or not refraction is enabled for the observer line of sight rays. Requires
        the refractive index to be set in the Geometry object for refraction to work.  Defaults to False.
        """
        self._config.los_refraction = value

    @property
    def solar_refraction(self) -> bool:
        """
        Controls whether or not refraction is enabled for the solar rays. Requires
        the refractive index to be set in the Geometry object for refraction to work.  Defaults to False.
        """
        return self._config.solar_refraction

    @solar_refraction.setter
    def solar_refraction(self, value: bool):
        """
        Controls whether or not refraction is enabled for the solar rays. Requires
        the refractive index to be set in the Geometry object for refraction to work.  Defaults to False.
        """
        self._config.solar_refraction = value

    @property
    def output_los_optical_depth(self) -> bool:
        """
        If True, then the output from a calculate_radiance call will also contain an "optical_depth" variable
        with dimensions ["los", "wavelength"] that is the optical depth for each line of sight.
        """
        return self._config.output_los_optical_depth

    @output_los_optical_depth.setter
    def output_los_optical_depth(self, value: bool):
        """
        If True, then the output from a calculate_radiance call will also contain an "optical_depth" variable
        with dimensions ["los", "wavelength"] that is the optical depth for each line of sight.
        """
        self._config.output_los_optical_depth = value

    @property
    def multiple_scatter_refraction(self) -> bool:
        """
        Controls whether or not refraction is enabled for the solar line of sight rays. Requires
        the refractive index to be set in the Geometry object for refraction to work.  Only has an effect
        when the single scatter source term is set to Table.  Defaults to False.
        """
        return self._config.multiple_scatter_refraction

    @multiple_scatter_refraction.setter
    def multiple_scatter_refraction(self, value: bool):
        """
        Controls whether or not refraction is enabled for the solar line of sight rays. Requires
        the refractive index to be set in the Geometry object for refraction to work.  Only has an effect
        when the single scatter source term is set to Table.  Defaults to False.
        """
        self._config.multiple_scatter_refraction = value

    @property
    def num_sza(self) -> int:
        """
        The number of solar zenith angle discretizations to use when calculating the multiple scatter source.
        For the discrete ordinates source, this determines the number of independent discrete ordinates calculations to perform.
        In the successive orders of scattering source, this is directly the number of discretizations.
        Defaults to 1, indicating that the multiple scatter source is estimated only at the reference point.
        """
        return self._config.num_sza

    @num_sza.setter
    def num_sza(self, value: int):
        """
        The number of solar zenith angle discretizations to use when calculating the multiple scatter source.
        For the discrete ordinates source, this determines the number of independent discrete ordinates calculations to perform.
        In the successive orders of scattering source, this is directly the number of discretizations.
        Defaults to 1, indicating that the multiple scatter source is estimated only at the reference point.
        """
        self._config.num_sza = value

    @property
    def num_successive_orders_iterations(self) -> int:
        """
        The number of iterations to perform when using the successive orders of scattering multiple scatter
        source.
        """
        return self._config.num_successive_orders_iterations

    @num_successive_orders_iterations.setter
    def num_successive_orders_iterations(self, value: int):
        """
        The number of iterations to perform when using the successive orders of scattering multiple scatter
        source.
        """
        self._config.num_successive_orders_iterations = value

    @property
    def init_successive_orders_with_discrete_ordinates(self) -> bool:
        """
        If set to true, when using the successive orders source, it will be initialized with a source calculated with
        the discrete ordinates source instead of a single scattering source.  This greatly reduces the number
        of iterations required for the method, as well as provides a better weighting function approximation.
        """
        return self._config.init_successive_orders_with_discrete_ordinates

    @init_successive_orders_with_discrete_ordinates.setter
    def init_successive_orders_with_discrete_ordinates(self, value: bool):
        """
        If set to true, when using the successive orders source, it will be initialized with a source calculated with
        the discrete ordinates source instead of a single scattering source.  This greatly reduces the number
        of iterations required for the method, as well as provides a better weighting function approximation.
        """
        self._config.init_successive_orders_with_discrete_ordinates = value

    @property
    def num_streams(self) -> int:
        """
        The number of streams to use in the discrete ordinates method. This is the number of streams
        in full space, i.e. each hemisphere has num_streams / 2 angular discretizations.  Must
        be an even number. Default to 16.
        """
        return self._config.num_streams

    @num_streams.setter
    def num_streams(self, value: int):
        """
        The number of streams to use in the discrete ordinates method. This is the number of streams
        in full space, i.e. each hemisphere has num_streams / 2 angular discretizations.  Must
        be an even number. Default to 16.
        """
        self._config.num_streams = value

    @property
    def num_forced_azimuth(self) -> int:
        """
        If set to a value greater than 0, the discrete ordinates method will use this number of azimuth terms independent of convergence.
        Defaults to -1, which means to use the number of azimuth terms required for convergence.
        """
        return self._config.num_forced_azimuth

    @num_forced_azimuth.setter
    def num_forced_azimuth(self, value: int):
        """
        If set to a value greater than 0, the discrete ordinates method will use this number of azimuth terms independent of convergence.
        Defaults to -1, which means to use the number of azimuth terms required for convergence.
        """
        self._config.num_forced_azimuth = value

    @property
    def do_backprop(self) -> bool:
        """
        Enables backpropagation for the weighting functions when using the DO source in plane parallel or pseudo-spherical geometry.
        Can greatly improve the computation speed of the calculation when the number of lines of sight is small. Default to True
        """
        return self._config.do_backprop

    @do_backprop.setter
    def do_backprop(self, value: bool):
        """
        Enables backpropagation for the weighting functions when using the DO source in plane parallel or pseudo-spherical geometry.
        Can greatly improve the computation speed of the calculation when the number of lines of sight is small. Default to True
        """
        self._config.do_backprop = value

    @property
    def num_successive_order_points(self) -> int:
        """
        The number of incoming points to use in the successive orders calculation for each solar
        zenith angle.  Must be equal to or less than the number of atmosphere altitude grid points.
        Default is -1 which means to use every altitude grid point.
        """
        return self._config.num_successive_order_points

    @num_successive_order_points.setter
    def num_successive_order_points(self, value: int):
        """
        The number of incoming points to use in the successive orders calculation for each solar
        zenith angle.  Must be equal to or less than the number of atmosphere altitude grid points.
        Default is -1 which means to use every altitude grid point.
        """
        self._config.num_successive_order_points = value

    @property
    def num_singlescatter_moments(self) -> int:
        """
        The number of Legendre expansion moments to use in the single scatter calculation.
        Must be greater or equal to num_streams. Default to 16.
        """
        return self._config.num_singlescatter_moments

    @num_singlescatter_moments.setter
    def num_singlescatter_moments(self, value: int):
        """
        The number of Legendre expansion moments to use in the single scatter calculation.
        Must be greater or equal to num_streams. Default to 16.
        """
        self._config.num_singlescatter_moments = value

    @property
    def num_successive_orders_incoming(self) -> int:
        """
        The number of integration nodes to use in the successive orders algorithm when calculating the incoming
        radiance at each grid point.  Must be one of [6, 14, 26, 38, 50, 74, 86, 110, 146, 170, 194, 230, 266, 302, 350,
        434, 590, 770, 974, 1202, 1454, 1730, 2030, 2354, 2702, 3074, 3470, 3890]. Default is 110.
        """
        return self._config.num_successive_orders_incoming

    @num_successive_orders_incoming.setter
    def num_successive_orders_incoming(self, value: int):
        """
        The number of integration nodes to use in the successive orders algorithm when calculating the incoming
        radiance at each grid point.  Must be one of [6, 14, 26, 38, 50, 74, 86, 110, 146, 170, 194, 230, 266, 302, 350,
        434, 590, 770, 974, 1202, 1454, 1730, 2030, 2354, 2702, 3074, 3470, 3890]. Default is 110.
        """
        self._config.num_successive_orders_incoming = value

    @property
    def num_successive_orders_outgoing(self) -> int:
        """
        The number of sample points to use in the successive orders algorithm to calculate the outgoing source function on
        radiance at each grid point.  Must be one of [6, 14, 26, 38, 50, 74, 86, 110, 146, 170, 194, 230, 266, 302, 350,
        434, 590, 770, 974, 1202, 1454, 1730, 2030, 2354, 2702, 3074, 3470, 3890]. Default is 110.
        """
        return self._config.num_successive_orders_outgoing

    @num_successive_orders_outgoing.setter
    def num_successive_orders_outgoing(self, value: int):
        """
        The number of sample points to use in the successive orders algorithm to calculate the outgoing source function on
        radiance at each grid point.  Must be one of [6, 14, 26, 38, 50, 74, 86, 110, 146, 170, 194, 230, 266, 302, 350,
        434, 590, 770, 974, 1202, 1454, 1730, 2030, 2354, 2702, 3074, 3470, 3890]. Default is 110.
        """
        self._config.num_successive_orders_outgoing = value

    @property
    def threading_lib(self) -> ThreadingLib:
        """
        Sets the threading library to use in the calculation.  Default is ThreadingLib.Rayon.
        """
        return self._config.threading_lib

    @threading_lib.setter
    def threading_lib(self, value: ThreadingLib):
        """
        Sets the threading library to use in the calculation.  Default is ThreadingLib.Rayon.
        """
        self._config.threading_lib = value



from __future__ import annotations

from sasktran2._core_rust import (
    PyGroundViewingSolar,
    PySolarAnglesObserverLocation,
    PyTangentAltitudeSolar,
    PyViewingGeometry,
)


class ViewingGeometry:
    _viewing_geometry: PyViewingGeometry

    def __init__(self):
        self._viewing_geometry = PyViewingGeometry()
        self._rays = []

    def add_ray(self, ray):
        self._rays.append(ray)
        self._viewing_geometry.add_ray(ray._internal)

    @property
    def observer_rays(self):
        return self._rays


class TangentAltitudeSolar:
    _internal: PyTangentAltitudeSolar

    def __init__(
        self,
        tangent_altitude_m: float,
        relative_azimuth: float,
        observer_altitude_m: float,
        cos_sza: float,
    ):
        """
        Defines a viewing ray from the observer altitude, and tangent point parameters. Note that all of
        these parameters assume straight line paths (i.e. no atmospheric refraction)

        Parameters
        ----------
        tangent_altitude_m: float
            Tangent altitude in [m]
        relative_azimuth: float
            Relative azimuth angle to the sun. An angle of 0 degrees corresponds to the forward scattering plane. [rad]
        observer_altitude_m: float
            Observer altitude relative to the earth [m]
        cos_sza: float
            Cosine of the solar zenith angle at the tangent point [unitless]
        """
        self._internal = PyTangentAltitudeSolar(
            tangent_altitude_m, relative_azimuth, observer_altitude_m, cos_sza
        )

        self._tangent_altitude_m = tangent_altitude_m
        self._relative_azimuth = relative_azimuth
        self._observer_altitude_m = observer_altitude_m
        self._cos_sza = cos_sza

    def __repr__(self):
        # "Tangent Viewing Ray: tangentaltitude: {}, relative_azimuth_angle: "
        #    "{}, observeraltitude: {}, theta: {}, phi: {}",

        return f"Tangent Viewing Ray: tangentaltitude: {self._tangent_altitude_m}, relative_azimuth_angle: {self._relative_azimuth}, observeraltitude: {self._observer_altitude_m}, cos_sza: {self._cos_sza}"


class GroundViewingSolar:
    _internal: PyGroundViewingSolar

    def __init__(
        self,
        cos_sza: float,
        relative_azimuth: float,
        cos_viewing_zenith: float,
        observer_altitude_m: float,
    ):
        """
                Defines a viewing ray that is looking at the ground from angles defined at the ground location. Note that
        all of these parameters assumes straight line paths (i.e. no atmospheric refraction)

        Parameters
        ----------
        cos_sza: float
            Cosine of solar zenith angle at the ground point [unitless]
        relative_azimuth: float
            Relative azimuth angle to the sun [rad] at the ground point. An angle of 0 degrees corresponds to the forward scattering plane.
        observer_altitude_m: float
            Observer altitude relative to the earth [m]
        cos_viewing_zenith: float
            Cosine of the viewing zenith angle at the ground point [unitless]
        """
        self._internal = PyGroundViewingSolar(
            cos_sza, relative_azimuth, cos_viewing_zenith, observer_altitude_m
        )

        self._cos_sza = cos_sza
        self._relative_azimuth = relative_azimuth
        self._cos_viewing_zenith = cos_viewing_zenith
        self._observer_altitude_m = observer_altitude_m

    def __repr__(self):
        #            "Ground Viewing Ray: cos_sza: {}, relative_azimuth_angle: {}, "
        #    "cos_viewing_zenith: {}, observer_altitude: {}",
        return f"Ground Viewing Ray: cos_sza: {self._cos_sza}, relative_azimuth_angle: {self._relative_azimuth}, cos_viewing_zenith: {self._cos_viewing_zenith}, observer_altitude_m: {self._observer_altitude_m}"


class SolarAnglesObserverLocation:
    _internal: PySolarAnglesObserverLocation

    def __init__(
        self,
        cos_sza: float,
        relative_azimuth: float,
        cos_viewing_zenith: float,
        observer_altitude_m: float,
    ):
        """
        Defines a viewing ray that is defined at a location defined from the solar angles. Note that
        all of these parameters assumes straight line paths (i.e. no atmospheric refraction).
        This differs from sk.GroundViewingSolar in that the angles are defined at the observer location, not the ground location.

        Parameters
        ----------
        cos_sza: float
            Cosine of solar zenith angle at the observer point [unitless]
        relative_azimuth: float
            Relative azimuth angle to the sun [rad] at the observer point. An angle of 0 degrees corresponds to the forward scattering plane.
        cos_viewing_zenith: float
            Cosine of the viewing zenith angle at the observer point.  Positive angles are viewing up,
            negative angles are viewing down. [unitless]
        observer_altitude_m: float
            Observer altitude relative to the earth [m]
        """
        self._internal = PySolarAnglesObserverLocation(
            cos_sza, relative_azimuth, cos_viewing_zenith, observer_altitude_m
        )

        self._cos_sza = cos_sza
        self._relative_azimuth = relative_azimuth
        self._cos_viewing_zenith = cos_viewing_zenith
        self._observer_altitude_m = observer_altitude_m

    def __repr__(self):
        #            "Up Viewing Ray: cos_sza: {}, relative_azimuth_angle: {}, "
        #    "cos_viewing_zenith: {}, observer_altitude: {}",
        return f"Up Viewing Ray: cos_sza: {self._cos_sza}, relative_azimuth_angle: {self._relative_azimuth}, cos_viewing_zenith: {self._cos_viewing_zenith}, observer_altitude: {self._observer_altitude_m}"



from __future__ import annotations

import logging
from copy import copy
from dataclasses import dataclass

import numpy as np

import sasktran2 as sk
from sasktran2._core_rust import EmissionSource, PyAtmosphere
from sasktran2.polarization import LegendreStorageView
from sasktran2.units import (
    wavenumber_cminv_to_wavlength_nm,
    wavlength_nm_to_wavenumber_cminv,
)
from sasktran2.util.state import EquationOfState


@dataclass
class NativeGridDerivative:
    """
    Internal input object that defines the model input quantities necessary to calculate a derivative.
    This mapping is from the native model grid to the native grid, it does not change the gridding.
    """

    d_extinction: np.ndarray = None
    d_ssa: np.ndarray = None
    d_leg_coeff: np.ndarray = None
    scat_factor: np.ndarray = None
    scat_deriv_index: int = None
    d_brdf: np.ndarray = None


class Atmosphere:
    def __init__(
        self,
        model_geometry: sk.Geometry1D,
        config: sk.Config,
        wavelengths_nm: np.array = None,
        wavenumber_cminv: np.array = None,
        numwavel: int | None = None,
        calculate_derivatives: bool = True,
        pressure_derivative: bool = True,
        temperature_derivative: bool = True,
        specific_humidity_derivative: bool = True,
        legendre_derivative: bool = True,
    ):
        """
        The main specification for the atmospheric state.

        See :py:attr:`~storage` for details on how the atmospheric state parameters are stored.

        See :py:attr:`~surface` for details on how the surface parameters are stored.

        Parameters
        ----------
        model_geometry : sk.Geometry1D
            The geometry defining where the atmospheric quantities are specified on.
        config : sk.Config
            Main configuration object.
        wavelengths_nm : np.array, optional
            Array of wavelengths to specify the atmosphere at in [nm].  One of wavelengths_nm, wavenumber_cminv, numwavel must be set.
        wavenumber_cminv : np.array, optional
            Array of wavenumbers to specify the atmosphere at in [:math:`\\text{cm}^{-1}`]. One of wavelengths_nm, wavenumber_cminv, numwavel must be set.
        numwavel : int, optional
            Number of wavelengths to include in the calculation.  Note that wavelength is a dummy variable, this dimension can be
            anaything. One of wavelengths_nm, wavenumber_cminv, numwavel must be set.
        calculate_derivatives : bool, optional
            Whether or not the model should calculate derivatives with respect to atmospheric quantities., by default True
        pressure_derivative: bool, optional
            Whether or not the model should calculate derivatives with respect to pressure., by default True
        temperature_derivative: bool, optional
            Whether or not the model should calculate derivatives with respect to temperature., by default True
        legendre_derivative: bool, optional
            Whether or not the model should calculate derivatives with respect to the legendre coefficients., by default True
        """
        self._wavelengths_nm = None
        self._wavenumbers_cminv = None

        # Set wavelengths
        if (
            (wavelengths_nm is None)
            and (wavenumber_cminv is None)
            and (numwavel is None)
        ):
            msg = "One of wavelengths_nm, wavenumber_cminv, numwavel must be set when constructing the Atmosphere object"
            raise ValueError(msg)

        if wavelengths_nm is not None:
            self.wavelengths_nm = wavelengths_nm.astype(np.float64)

        if wavenumber_cminv is not None:
            self.wavenumbers_cminv = wavenumber_cminv.astype(np.float64)

        nwavel = len(self.wavelengths_nm) if numwavel is None else numwavel

        self._nstokes = config.num_stokes
        self._calculate_derivatives = calculate_derivatives
        self._pressure_derivative = pressure_derivative
        self._temperature_derivative = temperature_derivative
        self._legendre_derivative = legendre_derivative
        self._specific_humidity_derivative = specific_humidity_derivative

        self._atmosphere = PyAtmosphere(
            nwavel,
            len(model_geometry.altitudes()),
            config.num_singlescatter_moments,
            calculate_derivatives,
            config.emission_source != EmissionSource.NoSource,
            config.num_stokes,
        )

        self._leg_coeff = LegendreStorageView(
            self._atmosphere.storage.leg_coeff, self._nstokes
        )

        self._storage_needs_reset = False

        self._equation_of_state = EquationOfState()

        self._model_geometry = model_geometry
        self._config = config

        self._constituents = {}
        self._derivs = {}
        self._unscaled_ssa = None
        self._unscaled_extinction = None
        self._applied_delta_m_order = None
        self._nwavel = nwavel

    @property
    def num_wavel(self) -> int:
        """
        The number of wavelengths the atmosphere is specified at

        Returns
        -------
        int
        """
        return self._nwavel

    @property
    def model_geometry(self) -> sk.Geometry1D:
        """
        The model geometry object

        Returns
        -------
        sk.Geometry1D
        """
        return self._model_geometry

    @property
    def applied_delta_m_order(self) -> int | None:
        """
        The order of the applied delta_m scaling. Can be None if no scaling has been applied
        """
        return self._applied_delta_m_order

    @property
    def nstokes(self) -> int:
        """
        The number of stokes parameters the atmosphere contains information to calculate

        Returns
        -------
        int
        """
        return self._nstokes

    @property
    def calculate_temperature_derivative(self) -> bool:
        """
        True if we are calculating the derivative with respect to temperature

        Returns
        -------
        bool
        """
        return self._temperature_derivative

    @property
    def calculate_pressure_derivative(self) -> bool:
        """
        True if we are calculating the derivative with respect to pressure

        Returns
        -------
        bool
        """
        return self._pressure_derivative

    @property
    def calculate_specific_humidity_derivative(self) -> bool:
        """
        True if we are calculating the derivative with respect to specific humidity

        Returns
        -------
        bool
        """
        return self._specific_humidity_derivative

    @property
    def storage(
        self,
    ) -> sk.AtmosphereStorageStokes_1 | sk.AtmosphereStorageStokes_3:
        """
        The internal object which contains the atmosphere extinction, single scatter albedo, and legendre coefficients.

        Returns
        -------
        Union[sk.AtmosphereGridStorageStokes_1, sk.AtmosphereGridStorageStokes_3]
        """
        return self._atmosphere.storage

    @property
    def surface(self) -> sk.SurfaceStokes_1 | sk.SurfaceStokes_3:
        """
        The surface object

        Returns
        -------
        sk.Surface
        """
        return self._atmosphere.surface

    @property
    def temperature_k(self) -> np.array:
        """
        Atmospheric temperature in [K] on the same grid as :py:attr:`~model_geometry`

        Returns
        -------
        np.array
        """
        return self._equation_of_state.temperature_k

    @temperature_k.setter
    def temperature_k(self, temp: np.array):
        self._equation_of_state.temperature_k = temp

    @property
    def pressure_pa(self) -> np.array:
        """
        Pressure in [Pa] on the same grid as :py:attr:`~model_geometry`

        Returns
        -------
        np.array
        """
        return self._equation_of_state.pressure_pa

    @property
    def specific_humidity(self) -> np.array:
        """
        Specific humidity on the same grid as :py:attr:`~model_geometry`

        Returns
        -------
        np.array
        """
        return self._equation_of_state.specific_humidity

    @property
    def state_equation(self) -> EquationOfState:
        """
        The equation of state object which contains the temperature, pressure, and specific humidity

        Returns
        -------
        EquationOfState
        """
        return self._equation_of_state

    @specific_humidity.setter
    def specific_humidity(self, sh: np.array):
        self._equation_of_state.specific_humidity = sh

    @pressure_pa.setter
    def pressure_pa(self, pres: np.array):
        self._equation_of_state.pressure_pa = pres

    @property
    def wavelengths_nm(self) -> np.ndarray | None:
        """
        The wavelengths in [nm] the atmosphere is specified at.  This is an
        optional property, it may be None.

        Returns
        -------
        Optional[np.array]
        """
        return self._wavelengths_nm

    @wavelengths_nm.setter
    def wavelengths_nm(self, wav: np.array):
        self._wavelengths_nm = wav
        self._wavenumbers_cminv = wavlength_nm_to_wavenumber_cminv(wav)

    @property
    def wavenumbers_cminv(self) -> np.ndarray | None:
        """
        The wavenumbers in [:math:`\\text{cm}^{-1}`].  This is an optional property, it may be set to None

        Returns
        -------
        Optional[np.array]
        """
        return self._wavenumbers_cminv

    @wavenumbers_cminv.setter
    def wavenumbers_cminv(self, wav: np.array):
        self._wavenumbers_cminv = wav
        self._wavelengths_nm = wavenumber_cminv_to_wavlength_nm(wav)

    def _zero_storage(self):
        """
        Sets the internal storage object to 0.  Typically used in-between calculations to reset the
        atmosphere without reconstructing it.
        """
        self.storage.set_zero()
        self.surface.set_zero()

    @property
    def deriv_mappings(self) -> dict:
        """
        A nested dictionary of :py:class:`sasktran2.atmosphere.DerivativeMapping` objects.

        Returns
        -------
        dict
        """
        return self._derivs

    @property
    def unscaled_ssa(self) -> np.ndarray:
        """
        The unscaled single scatter albedo.  After performing the calculation, `storage.ssa` may be
        modified because of the delta-m scaling.  This property stores the original unscaled single
        scatter albedo.

        Returns
        -------
        np.ndarray
        """
        return self._unscaled_ssa

    @property
    def unscaled_extinction(self) -> np.ndarray:
        """
        The unscaled extinction. After performing the calculation, `storage.total_extinction` may be
        modified because of the delta-m scaling. This property stores the original unscaled single
        scatter albedo.

        Returns
        -------
        np.ndarray
        """
        return self._unscaled_extinction

    @property
    def leg_coeff(self) -> LegendreStorageView:
        return self._leg_coeff

    def internal_object(self) -> sk.AtmosphereStokes_1 | sk.AtmosphereStokes_3:
        """
        The internal `pybind11` object that can be used to perform the radiative transfer calculation.
        Calling this method will trigger a construction of the atmosphere object if necessary, and then
        return back the internal object.

        Returns
        -------
        Union[sk.AtmosphereStokes_1, sk.AtmosphereStokes_3]
        """

        if len(self._constituents) > 0:
            logging.debug("Setting atmosphere from constituents")
            # Using the constituent interface
            if self._storage_needs_reset:
                self._zero_storage()
            for _, constituent in self._constituents.items():
                constituent.add_to_atmosphere(self)

            self.storage.normalize_by_extinctions()

            self._derivs = {}
            if self._calculate_derivatives:
                for name, constituent in self._constituents.items():
                    constituent.register_derivative(self, name)

            logging.debug("Finished setting atmosphere from constituents")
        else:
            # using the raw interface
            if self._calculate_derivatives and len(self._derivs) == 0:
                deriv_mapping = self.storage.get_derivative_mapping("wf_extinction")
                deriv_mapping.d_extinction[:] = 1
                deriv_mapping.d_ssa[:] = 0.0
                deriv_mapping.interp_dim = "altitude"

                deriv_mapping = self.storage.get_derivative_mapping("wf_ssa")
                deriv_mapping.d_extinction[:] = 0.0
                deriv_mapping.d_ssa[:] = 1.0
                deriv_mapping.interp_dim = "altitude"

                deriv_mapping = self.surface.get_derivative_mapping("wf_albedo")
                deriv_mapping.d_brdf[:] = 1.0

                if self._legendre_derivative:
                    for i in range(self.storage.leg_coeff.shape[0]):
                        if i == 0:
                            # No derivative for the first leg_coeff
                            continue

                        deriv_mapping = self.storage.get_derivative_mapping(
                            f"wf_leg_coeff_{i}"
                        )
                        deriv_mapping.d_leg_coeff[i, :] = 1
                        deriv_mapping.d_extinction[:] = 0
                        deriv_mapping.d_ssa[:] = 0
                        deriv_mapping.scat_factor[:] = 1
                        deriv_mapping.interp_dim = "altitude"

        # Now we need to resize the phase derivative storage if necessary, and set the scattering derivatives
        self.storage.finalize_scattering_derivatives(0)

        # Store the unscaled optical properties for use in the derivative mappings
        self._unscaled_ssa = copy(self.storage.ssa)
        self._unscaled_extinction = copy(self.storage.total_extinction)

        # Apply delta scaling if required
        if self._config.delta_m_scaling:
            # Find the order of the scaling
            if self._config.num_streams == self._leg_coeff.a1.shape[0]:
                # Can't apply scaling
                logging.info(
                    "Delta-m Scaling NOT applied since the number of MS streams is equal to the number of user input legendre coefficients"
                )
            else:
                self._atmosphere.apply_delta_m_scaling(self._config.num_streams)
                self._applied_delta_m_order = self._config.num_streams

        self._storage_needs_reset = True
        return self._atmosphere

    def __setitem__(self, item, value):
        self._constituents[item] = value

    def __getitem__(self, item):
        return self._constituents.get(item)



from __future__ import annotations

import shlex
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from sasktran2.database.base import CachedDatabase
from sasktran2.util import get_hapi


def _read_line_file_pd(file_path: Path) -> xr.Dataset:
    """
    Reads an AER linefile and puts it in the same format as the HITRAN optical reader.
    This function uses the pandas backend, it is faster but does not support all features

    Parameters
    ----------
    file_path : Path
        Path to the individual molecule datafile

    Returns
    -------
    xr.Dataset
    """
    # HITRAN F100 format is
    widths = np.array([2, 1, 12, 10, 10, 5, 5, 10, 4, 8, 3, 3, 9, 9, 3, 6])
    names = [
        "molec_id",
        "local_iso_id",
        "nu",
        "sw",
        "rsq",
        "gamma_air",
        "gamma_self",
        "elower",
        "n_air",
        "delta_air",
        "upper_quanta",
        "lower_quanta",
        "upper_local_q",
        "lower_local_q",
        "error_codes",
        "reference",
    ]

    data = pd.read_fwf(
        file_path,
        widths=widths,
        names=names,
        # converters=converters,
        comment=">",
        on_bad_lines="skip",
        error_bad_lines=False,
    )
    # Drop rows where mol_id is %%
    data = data[data["molec_id"] != "%%"]

    # Rows where iso_id is NaN are extra information
    data = data.dropna(subset=["local_iso_id"])
    data = data[data["local_iso_id"] != "-"]

    # In Line intensity we have to replace the D with an E, if it was loaded in as a float
    if data["sw"].dtype != float:
        data.loc[:, "sw"] = data["sw"].str.replace("D", "E")

    # Now perform the conversions
    data["molec_id"] = data["molec_id"].astype(int)
    data["local_iso_id"] = data["local_iso_id"].astype(int)
    data["nu"] = data["nu"].astype(float)
    data["sw"] = data["sw"].astype(float)
    data["rsq"] = data["rsq"].astype(float)
    data["gamma_air"] = data["gamma_air"].astype(float)
    data["gamma_self"] = data["gamma_self"].astype(float)
    data["elower"] = data["elower"].astype(float)
    data["n_air"] = data["n_air"].astype(float)
    data["delta_air"] = data["delta_air"].astype(float)

    # Drop all other columns
    data = data[
        [
            "molec_id",
            "local_iso_id",
            "nu",
            "sw",
            "rsq",
            "gamma_air",
            "gamma_self",
            "elower",
            "n_air",
            "delta_air",
        ]
    ]

    # Reset the index column
    data = data.reset_index(drop=True)

    # TODO: Speed dependent parameters, line coupling, and line mixing
    return xr.Dataset.from_dataframe(data)


def _read_line_file_py(file_path: Path) -> xr.Dataset:
    """
    Reads an AER linefile and puts it in the same format as the HITRAN optical reader.
    This function uses the python backend, it is slower but supports all features such as
    line coupling

    Parameters
    ----------
    file_path : Path
        Path to the individual molecule datafile

    Returns
    -------
    xr.Dataset
    """
    # HITRAN F100 format is
    widths = np.array([2, 1, 12, 10, 10, 5, 5, 10, 4, 8, 3, 3, 9, 9, 3, 6])
    cols = np.concatenate(([0], np.cumsum(widths)))  # e.g. [ 0,  2,  3, 15, 25, ...]

    names = {
        "molec_id": int,
        "local_iso_id": int,
        "nu": float,
        "sw": float,
        "rsq": float,
        "gamma_air": float,
        "gamma_self": float,
        "elower": float,
        "n_air": float,
        "delta_air": float,
        "upper_quanta": str,
        "lower_quanta": str,
        "upper_local_q": str,
        "lower_local_q": str,
        "error_codes": str,
        "reference": str,
    }
    name_items = list(names.items())  # for indexing

    all_data = []

    with file_path.open() as f:
        for line in f:
            # Skip blank lines or comment lines
            if (not line.strip()) or line[0] in (">", "%"):
                continue

            # Slice out all fields at once
            fields = [
                line[cols[i] : cols[i + 1]].replace("D", "E")
                for i in range(len(widths))
            ]

            # Parse them according to the dictionary
            record = {}
            for (k, cast), raw in zip(name_items, fields, strict=True):
                record[k] = cast(raw)

            # Check if we need to read the line-coupling parameters
            if record["reference"][-2:-1] == "-":
                lc = next(f)  # read the next line
                record["Y_200"] = float(lc[3:15])
                record["G_200"] = float(lc[15:26])
                record["Y_250"] = float(lc[26:39])
                record["G_250"] = float(lc[39:50])
                record["Y_296"] = float(lc[50:63])
                record["G_296"] = float(lc[63:74])
                record["Y_340"] = float(lc[74:87])
                record["G_340"] = float(lc[87:98])
            else:
                record["Y_200"] = np.nan
                record["G_200"] = np.nan
                record["Y_250"] = np.nan
                record["G_250"] = np.nan
                record["Y_296"] = np.nan
                record["G_296"] = np.nan
                record["Y_340"] = np.nan
                record["G_340"] = np.nan

            all_data.append(record)

    # Convert to dataframe
    data = pd.DataFrame(all_data)

    # Reset the index column
    data = data.reset_index(drop=True)

    ds = xr.Dataset.from_dataframe(data)

    ds["Y"] = ds[["Y_200", "Y_250", "Y_296", "Y_340"]].to_array(dim="temperature")
    ds["G"] = ds[["G_200", "G_250", "G_296", "G_340"]].to_array(dim="temperature")

    ds = ds.drop_vars(
        ["Y_200", "G_200", "Y_250", "G_250", "Y_296", "G_296", "Y_340", "G_340"]
    )

    ds = ds.assign_coords(temperature=[200.0, 250.0, 296.0, 340.0])

    return ds.drop(
        [
            "error_codes",
            "reference",
            "upper_quanta",
            "lower_quanta",
            "upper_local_q",
            "lower_local_q",
        ]
    )


class AERLineDatabase(CachedDatabase):
    """
    Database for AER line files.  Currently only supports the 3.8.1 version of the AER line database
    """

    def __init__(
        self,
        db_root: Path | None = None,
        rel_path: Path | None = "aer_lines",
        version: str = "3.8.1",
    ):
        CachedDatabase.__init__(self, db_root, rel_path=rel_path)

        self._hapi = get_hapi()
        self._version = version

        self._version_map = {"3.8.1": "5120012"}

    def path(self, _key: str, **kwargs) -> Path:
        dir = self._db_root
        version_dir = dir.joinpath(f"aer_v_{self._version}")
        # check if the AER database is downloaded, if not download it

        if not version_dir.exists():
            from zenodo_get import zenodo_get

            try:
                zenodo_get(
                    shlex.split(
                        f'--record {self._version_map[self._version]} -o "{dir.as_posix()}"'
                    )
                )
            except ImportError as e:
                msg = "zenodo_get is required to download the AER line database"
                raise ImportError(msg) from e

            file = dir.joinpath(f"aer_v_{self._version}.tar.gz")
            # Extract the tar file
            import tarfile

            with tarfile.open(file, "r:gz") as tar:
                tar.extractall(dir)
            # Delete the tar file
            file.unlink()

        return version_dir

    def load_ds(self, _key: str, **kwargs) -> xr.Dataset:
        return None

    def clear(self):
        # Delete the entire database folder
        folder = self._db_root.joinpath(f"aer_v_{self._version}")
        if folder.exists():
            for file in folder.glob("*"):
                if file.is_dir():
                    for subfile in file.glob("*"):
                        subfile.unlink()
                    file.rmdir()
                else:
                    file.unlink()



from __future__ import annotations

import numpy as np

from sasktran2._core_rust import PyVMRAltitudeAbsorber
from sasktran2.atmosphere import Atmosphere
from sasktran2.optical.base import OpticalProperty

from .base import Constituent


class VMRAltitudeAbsorber(Constituent):
    _constituent: PyVMRAltitudeAbsorber

    def __init__(
        self,
        optical_property: OpticalProperty,
        altitudes_m: np.array,
        vmr: np.array,
        out_of_bounds_mode: str = "zero",
    ) -> None:
        """
        An atmospheric constituent that is specified through volume mixing ratio (VMR) on an altitude grid.
        The altitude grid need not match the global atmospheric grid, in the case they are different the VMRs
        will be linearly interpolated to the atmospheric grid locations.

        The constituent has an associated OpticalProperty that defines the absorption cross section as a function
        of wavelength.

        Parameters
        ----------
        optical_property : OpticalProperty
            An object that provides absorption cross sections as a function of wavelength
        altitudes_m : np.array
            Altitudes in [m] that the VMR is specified on
        vmr : np.array
            Volume mixing ratio
        out_of_bounds_mode : str, optional
            One of ['zero', 'extend'], 'zero' will set the VMR to be 0 on any altitudes that are
            out of bounds.  'extend' will extend the last or first VMR value.  By default 'zero'
        """
        super().__init__()

        self._constituent = PyVMRAltitudeAbsorber(
            optical_property,
            altitudes_m.astype(np.float64),
            vmr.astype(np.float64),
            out_of_bounds_mode,
        )

        self._temp = optical_property

    @property
    def vmr(self):
        return self._constituent.vmr

    @vmr.setter
    def vmr(self, vmr: np.array):
        self._constituent.vmr = vmr

    def add_to_atmosphere(self, atmo: Atmosphere):
        self._constituent.add_to_atmosphere(atmo)

    def register_derivative(self, atmo: Atmosphere, name: str):
        self._constituent.register_derivative(atmo, name)



from __future__ import annotations

import numpy as np

from sasktran2._core_rust import GeometryType, InterpolationMethod, PyGeometry1D


class Geometry1D:
    _geometry: PyGeometry1D

    def __init__(
        self,
        cos_sza: float,
        solar_azimuth: float,
        earth_radius_m: float,
        altitude_grid_m: np.ndarray,
        interpolation_method: InterpolationMethod,
        geometry_type: GeometryType,
    ):
        self._geometry = PyGeometry1D(
            cos_sza,
            solar_azimuth,
            earth_radius_m,
            np.atleast_1d(altitude_grid_m).astype(np.float64),
            interpolation_method,
            geometry_type,
        )

    def altitudes(self) -> np.ndarray:
        return self._geometry.altitudes()

    @property
    def refractive_index(self) -> np.ndarray:
        return self._geometry.refractive_index

    @refractive_index.setter
    def refractive_index(self, value: np.ndarray) -> None:
        self._geometry.refractive_index = value



from __future__ import annotations

import numpy as np

from sasktran2._core_rust import PyMie, PyMieOutput


class MieValueAccessor:
    def __init__(self, owner):
        self._owner = owner

    def __getattr__(self, name):
        # Called when accessing "a.value.X"
        return getattr(self._owner, name)


class MieOutput:
    _mie_output: PyMieOutput

    def __init__(self, mie_output: PyMieOutput):
        self._mie_output = mie_output

    @property
    def values(self) -> MieValueAccessor:
        return MieValueAccessor(self)

    @property
    def cos_angles(self) -> np.ndarray:
        return self._mie_output.cos_angles

    @property
    def size_parameter(self) -> np.ndarray:
        return self._mie_output.size_param

    @property
    def Qsca(self) -> np.ndarray:
        return self._mie_output.Qsca

    @property
    def Qext(self) -> np.ndarray:
        return self._mie_output.Qext

    @property
    def S1(self) -> np.ndarray:
        return self._mie_output.S1

    @property
    def S2(self) -> np.ndarray:
        return self._mie_output.S2


class LinearizedMie:
    _mie: PyMie

    def __init__(self, num_threads: int = 1):  # noqa: ARG002
        self._mie = PyMie()

    def calculate(
        self,
        size_param: np.ndarray,
        refractive_index,
        cos_angles,
        calculate_derivatives=False,
    ):
        res = self._mie.calculate(
            np.atleast_1d(size_param).astype(np.float64),
            complex(refractive_index),
            np.atleast_1d(cos_angles).astype(np.float64),
            calculate_derivatives,
        )

        return MieOutput(res)



from __future__ import annotations

import logging

import numpy as np

from sasktran2._core_rust import (
    voigt_broaden,
    voigt_broaden_uniform,
    voigt_broaden_with_line_coupling,
)
from sasktran2.atmosphere import Atmosphere
from sasktran2.database.hitran_line import HITRANLineDatabase
from sasktran2.optical.base import OpticalProperty, OpticalQuantities
from sasktran2.util import get_hapi


class LineAbsorber(OpticalProperty):
    def __init__(
        self, molecule: str, line_db, line_fn="voigt", line_coupling=False, **kwargs
    ):
        """
        Absorption cross sections calculated using discrete absorption lines from a HITRAN-like database
        and broadened using the internal SASKTRAN2 Voigt line shape function.

        Notes:

        A line database is not included with SASKTRAN2 and must be downloaded separately. The hitran-api package
        is required to use HITRAN lines, and the zenodo_get package is required to download the AER line database.

        Broadening is calculated "online" rather than using pre-computed tables. This is slower but allows for
        more flexibility in the calculation.

        Derivatives with respect to pressure/temperature are currently not supported

        Parameters
        ----------
        molecule : str
            HITRAN Molecule identifier, e.g. "CO2", "H2O", "O3", "CH4"
        line_db : LineDatabase
            Line database to use, one of HITRANLineDatabase or AERLineDatabase
        line_fn : str
            Line shape function to use, currently only "voigt" is supported
        kwargs : dict
            Additional keyword arguments to pass to the line broadening function
        """
        defaults = {
            "line_contribution_width": 25.0,
            "cull_factor": 0.0,
            "subtract_pedastal": False,
        }

        self._line_db = line_db.load_ds(molecule)
        self._molecule = molecule
        self._kwargs = defaults
        self._kwargs.update(kwargs)
        self._line_coupling = line_coupling

        if line_fn != "voigt":
            msg = "Only Voigt line shape is supported"
            raise ValueError(msg)

        if self._molecule.lower() == "co2":
            # CO2 is bugged
            self._line_db.local_iso_id.values[
                self._line_db.local_iso_id.values == 0
            ] = 10

    def atmosphere_quantities(self, atmo: Atmosphere, **kwargs) -> OpticalQuantities:
        result = OpticalQuantities(
            extinction=np.zeros(
                (len(atmo.model_geometry.altitudes()), len(atmo.wavelengths_nm)),
                order="F",
            ),
            ssa=np.zeros(
                (len(atmo.model_geometry.altitudes()), len(atmo.wavelengths_nm)),
                order="F",
            ),
        )

        result.extinction[:] = self.cross_sections(
            atmo.wavelengths_nm,
            atmo.model_geometry.altitudes(),
            temperature_k=atmo.temperature_k,
            pressure_pa=atmo.pressure_pa,
            num_threads=atmo._config.num_threads,
            **kwargs,
        )
        return result

    def optical_derivatives(self, atmo: Atmosphere, **kwargs) -> dict:  # noqa: ARG002
        return {}

    def cross_sections(
        self, wavelengths_nm: np.array, altitudes_m: np.array, **kwargs  # noqa: ARG002
    ) -> OpticalQuantities:
        if "temperature_k" not in kwargs:
            msg = "Temperature must be provided to calculate the cross sections"
            raise ValueError(msg)
        if "pressure_pa" not in kwargs:
            msg = "Pressure must be provided to calculate the cross sections"
            raise ValueError(msg)

        num_threads = kwargs.get("num_threads", 1)

        hapi = get_hapi()

        wavenumbers_cminv = 1e7 / wavelengths_nm
        temperature_k = kwargs["temperature_k"]
        pressure_pa = kwargs["pressure_pa"]

        iso_keys = np.unique(self._line_db.local_iso_id.to_numpy())

        partition_ratio = np.zeros((len(temperature_k), len(iso_keys)))
        molecular_mass = np.zeros(len(iso_keys))

        for key in iso_keys:
            vals = hapi.partitionSum(
                self._line_db["molec_id"].to_numpy()[0],
                key,
                [*list(temperature_k), 296],
            )
            partition_ratio[:, key - 1] = vals[:-1] / vals[-1]
            molecular_mass[key - 1] = hapi.molecularMass(
                self._line_db["molec_id"].to_numpy()[0], key
            )

        sidx = np.argsort(wavenumbers_cminv)

        result = np.zeros(
            (len(pressure_pa), (len(wavenumbers_cminv))),
        )

        partial_pressures = kwargs.get("vmr", np.zeros(len(pressure_pa))) * pressure_pa
        logging.debug(f"Starting Broadening for {self._molecule}")

        if self._line_coupling:
            Y = (
                self._line_db["Y"]
                .fillna(0.0)
                .interp(temperature=temperature_k, kwargs={"fill_value": 0.0})
            )
            G = (
                self._line_db["G"]
                .fillna(0.0)
                .interp(temperature=temperature_k, kwargs={"fill_value": 0.0})
            )

            Y_arr = np.zeros((len(self._line_db.nu), len(temperature_k)))
            G_arr = np.zeros((len(self._line_db.nu), len(temperature_k)))
            Y_arr[:, :] = Y.to_numpy().T
            G_arr[:, :] = G.to_numpy().T

            voigt_broaden_with_line_coupling(
                self._line_db.nu.to_numpy(),
                self._line_db.sw.to_numpy(),
                self._line_db.elower.to_numpy(),
                self._line_db.gamma_air.to_numpy(),
                self._line_db.gamma_self.to_numpy(),
                self._line_db.delta_air.to_numpy(),
                self._line_db.n_air.to_numpy(),
                self._line_db.local_iso_id.to_numpy().astype(np.int32),
                partition_ratio,
                Y_arr,
                G_arr,
                molecular_mass,
                pressure_pa / 101325.0,
                np.ones(len(pressure_pa)) * 0,
                temperature_k,
                wavenumbers_cminv[sidx],
                result,
                num_threads=num_threads,
                **self._kwargs,
            )
            result[:, sidx] = result

            # Line coupling numerics can give very small negative values
            result[result < 0] = 0.0

            return result / 1e4

        # else
        # Check if the input is uniform wavenumber grid
        uniform_grid = np.allclose(
            np.diff(wavenumbers_cminv), np.diff(wavenumbers_cminv)[0], atol=1e-6
        )

        if uniform_grid:
            voigt_broaden_uniform(
                self._line_db.nu.to_numpy(),
                self._line_db.sw.to_numpy(),
                self._line_db.elower.to_numpy(),
                self._line_db.gamma_air.to_numpy(),
                self._line_db.gamma_self.to_numpy(),
                self._line_db.delta_air.to_numpy(),
                self._line_db.n_air.to_numpy(),
                self._line_db.local_iso_id.to_numpy().astype(np.int32),
                partition_ratio,
                molecular_mass,
                pressure_pa / 101325.0,
                partial_pressures / 101325.0,
                temperature_k.astype(np.float64),
                wavenumbers_cminv[sidx][0],
                wavenumbers_cminv[sidx][1] - wavenumbers_cminv[sidx][0],
                result,
                num_threads=num_threads,
                **self._kwargs,
            )

            result[:, sidx] = result

            return result / 1e4

        voigt_broaden(
            self._line_db.nu.to_numpy(),
            self._line_db.sw.to_numpy(),
            self._line_db.elower.to_numpy(),
            self._line_db.gamma_air.to_numpy(),
            self._line_db.gamma_self.to_numpy(),
            self._line_db.delta_air.to_numpy(),
            self._line_db.n_air.to_numpy(),
            self._line_db.local_iso_id.to_numpy().astype(np.int32),
            partition_ratio,
            molecular_mass,
            pressure_pa / 101325.0,
            np.ones(len(pressure_pa)) * 0,
            temperature_k,
            wavenumbers_cminv[sidx],
            result,
            num_threads=num_threads,
            **self._kwargs,
        )
        result[:, sidx] = result

        return result / 1e4

    def cross_section_derivatives(
        self, wavelengths_nm: np.array, altitudes_m: np.array, **kwargs  # noqa: ARG002
    ) -> dict:
        return {}


class HITRANAbsorber(LineAbsorber):
    def __init__(self, molecule: str, **kwargs):
        super().__init__(molecule, HITRANLineDatabase(), **kwargs)



from __future__ import annotations

from sasktran2._core_rust import openmp_support_enabled as py_openmp_support_enabled


def openmp_support_enabled() -> bool:
    """
    Check if OpenMP support is enabled in the current build of sasktran2.

    Returns:
        bool: True if OpenMP support is enabled, False otherwise.
    """
    return py_openmp_support_enabled()



from __future__ import annotations

import numpy as np

import sasktran2 as sk
from sasktran2._core_rust import PyRayleigh

from .base import Constituent


class Rayleigh(Constituent):
    _rayleigh: PyRayleigh

    def __init__(
        self,
        method: str = "bates",
        n2_percentage: float = 78.084,
        o2_percentage: float = 20.946,
        ar_percentage: float = 0.934,
        co2_percentage: float = 0.036,
        wavelengths_nm: np.ndarray = None,
        xs: np.ndarray = None,
        king_factor: np.ndarray = None,
    ) -> None:
        """
        An implementation of Rayleigh scattering.  Cross sections (and depolarization factors) can be
        calculated multiple ways, with the default method being that of 'bates'.

        Rayleigh scattering number density is estimated through the ideal gas law.

        This Constituent requires that the atmosphere object have `temperature_k`, `pressure_pa`, and
        `wavelength_nm` are all defined inside the :py:class:`sasktran2.Atmosphere` object.

        Parameters
        ----------
        method : str, default='bates'
            Method to use to calculate the cross section.  Supported methods are
            ['bates', 'manual'], by default 'bates'
        n2_percentage : float, optional
            Percentage of N2 in the atmosphere, by default 78.084
        o2_percentage : float, optional
            Percentage of O2 in the atmosphere, by default 20.946
        ar_percentage : float, optional
            Percentage of Ar in the atmosphere, by default 0.934
        co2_percentage : float, optional
            Percentage of CO2 in the atmosphere, by default 0.036
        wavelengths_nm : numpy.ndarray
            Wavelengths in nm to use for the cross section
        xs : numpy.ndarray
            Cross section in m2/molecule to use for the cross section
        king_factor : numpy.ndarray
            King factor to use for the cross section

        Raises
        ------
        ValueError
            If input method is not supported
        """
        self._rayleigh = PyRayleigh(
            method=method.lower(),
            n2_percentage=n2_percentage,
            o2_percentage=o2_percentage,
            ar_percentage=ar_percentage,
            co2_percentage=co2_percentage,
            wavelengths_nm=(
                wavelengths_nm.astype(np.float64)
                if wavelengths_nm is not None
                else None
            ),
            xs=xs.astype(np.float64) if xs is not None else None,
            king_factor=(
                king_factor.astype(np.float64) if king_factor is not None else None
            ),
        )

    def add_to_atmosphere(self, atmo: sk.Atmosphere):
        """
        Parameters
        ----------
        atmo : sk.Atmosphere


        :meta private:
        """
        self._rayleigh.add_to_atmosphere(atmo)

    def register_derivative(self, atmo: sk.Atmosphere, name: str):
        self._rayleigh.register_derivative(atmo, name)



from __future__ import annotations

import numpy as np

from sasktran2._core_rust import PyGeodetic


class Geodetic:
    _internal: PyGeodetic

    def __init__(self, radius: float, flattening: float):
        """
        A geodetic object that can be used to represent a location on the
        surface of the Earth. The geodetic object is based on the WGS84
        ellipsoid.

        Parameters
        ----------
        radius: float
            Radius of the ellipsoid in [m]
        flattening: float
            Flattening of the ellipsoid (1 - b/a)
        """
        self._internal = PyGeodetic(radius, flattening)

    @property
    def latitude(self) -> float:
        """
        Returns
        -------
        float
            Geodetic latitude in degrees
        """
        return self._internal.latitude

    @property
    def longitude(self) -> float:
        """
        Returns
        -------
        float
            Geodetic longitude in degrees
        """
        return self._internal.longitude

    @property
    def altitude(self) -> float:
        """
        Returns
        -------
        float
            Altitude in [m] above the surface of the ellipsoid.
        """
        return self._internal.altitude

    @property
    def valid(self) -> bool:
        """
        Returns
        -------
        bool
            True if the geodetic object has been initialized, False otherwise.
        """
        return self._internal.is_valid()

    def from_lat_lon_alt(self, latitude: float, longitude: float, altitude: float):
        """
        Initializes the Geodetic based on a specifiec latitude, longitude, and altitude.

        Parameters
        ----------
        latitude : float
            Latitude in degrees (-90 to 90)
        longitude : float
            Longitude in degrees (0 to 360 or -180 to 180)
        altitude : float
            Altitude above the geoid in metres

        Examples
        --------
        >>> import sasktran2 as sk
        >>> geodetic = sk.WGS84()
        >>> geodetic.from_lat_lon_alt(latitude=-15, longitude=-20, altitude=7342)
        >>> print(geodetic)
        WGS84 Location:
        Latitude: -15.0, Longitude: 340.0, Altitude: 7342.0
        """
        self._internal.from_lat_lon_alt(
            float(latitude), float(longitude), float(altitude)
        )

    def from_xyz(self, location: np.ndarray):
        """
        Initializes the Geodetic from a geocentric location

        Parameters
        ----------
        location : np.ndarray
            Three element vector containing a location in geocentric coordinates

        Examples
        --------
        >>> import sasktran2 as sk
        >>> geodetic = sk.WGS84()
        >>> geodetic.from_xyz([ 5797230.47518212, -2110019.3341472, -1642001.16317228])
        >>> print(geodetic)
        WGS84 Location:
        Latitude: -14.999999973747736, Longitude: 340.00000000000006, Altitude: 7344.999610390202
        """
        self._internal.from_xyz(np.atleast_1d(location).astype(np.float64))

    def from_tangent_altitude(
        self, altitude: float, observer: np.ndarray, boresight: np.ndarray
    ) -> np.ndarray:
        """
        Initialized the Geodetic from a specified tangent altitude, obsever location, and bore sight plane.

        Parameters
        ----------
        altitude : float
            Tangent altitude in meters
        observer : np.ndarray
            Three element array containing the obsever position in geocentric coordinates
        boresight : np.ndarray
            Three element array containing a normalized look vector that is within the bore sight plane.

        Returns
        -------
        np.ndarray
            Three element array containing the normalized look vector to the tangent point.

        Examples
        --------
        >>> import sasktran2 as sk
        >>> geodetic = sk.WGS84()
        >>> look = geodetic.from_tangent_altitude(15322, [ 3.676013154788849600e+005, 1.009976313640051500e+006,\
                                                            -6.871601202127538600e+006], [0, 0, 1])
        >>> print(look)
        [0.28880556 0.79348676 0.53569591]
        >>> print(geodetic)
        WGS84 Location:
        Latitude: -57.60888188776806, Longitude: 70.00000000000001, Altitude: 15321.971935882739
        """
        return self._internal.from_tangent_altitude(
            altitude,
            np.atleast_1d(observer).astype(np.float64),
            np.atleast_1d(boresight).astype(np.float64),
        )

    def from_tangent_point(self, observer, look_vector):
        """
        Initializes  the Geodetic by calculating the tangent point from an observer position and look vector

        Parameters
        ----------
        observer : np.ndarray
            Three element array containing the observer position in geocentric coordinates
        look_vector : np.ndarray
            Three element array containing a normalized look vector

        Examples
        --------
        >>> import sasktran2 as sk
        >>> geodetic = sk.WGS84()
        >>> geodetic.from_tangent_point([ 3.676013154788849600e+005, 1.009976313640051500e+006,\
                                            -6.871601202127538600e+006], [ 2.884568631765662100e-001,\
                                            7.925287180643269000e-001,  5.372996083468238900e-001])
        >>> print(geodetic)
        WGS84 Location:
        Latitude: -57.500000192733594, Longitude: 70.0, Altitude: 10002.99586173162
        """
        return self._internal.from_tangent_point(
            np.atleast_1d(observer).astype(np.float64),
            np.atleast_1d(look_vector).astype(np.float64),
        )

    @property
    def location(self) -> np.ndarray:
        """
        Returns
        -------
        np.ndarray
            Geocentric location in cartesian coordinates
        """
        return self._internal.location

    @property
    def local_south(self) -> np.ndarray:
        """
        Returns
        -------
        np.ndarray
            A unit vector pointing in the local south direction
        """
        return self._internal.local_south

    @property
    def local_up(self) -> np.ndarray:
        """
        Returns
        -------
        np.ndarray
            A unit vector pointing up (perpindicular to the ellipsoidal surface)
        """
        return self._internal.local_up

    @property
    def local_west(self) -> np.ndarray:
        """
        Returns
        -------
        np.ndarray
            A unit vector pointing in the local west direction
        """
        return self._internal.local_west

    def altitude_intercepts(
        self, altitude: float, observer: np.ndarray, look_vector: np.ndarray
    ) -> (np.ndarray, np.ndarray):  # type: ignore  # noqa: PGH003
        """
        Calculate the two intersections of a line of sight and an altitude.

        Parameters
        ----------
        altitude : float
            Altitude in meters.
        observer : np.ndarray
            Three element array containing the obsever position in geocentric coordinates.
        look_vector : np.ndarray
            Three element array containing a normalized look vector.

        Returns
        -------
        np.ndarray
            Three element array containing the first (entering) intercept in geocentric coordinates.
        np.ndarray
            Three element array containing the second (exiting) intercept in geocentric coordinates.

        Examples
        --------
        >>> import sasktran2 as sk
        >>> import numpy as np
        >>> geodetic = sk.WGS84()
        >>> look = geodetic.from_tangent_altitude(15322, [3.676013154788849600e+005, 1.009976313640051500e+006, \
                                                    -6.871601202127538600e+006], [0, 0, 1])
        >>> obs = geodetic.location
        >>> intercept1, intercept2 = geodetic.altitude_intercepts(16000, obs, look)
        >>> print(np.array_str(intercept1, precision=3))
        [ 1147302.059  3152186.5   -5425360.027]
        >>> print(np.array_str(intercept2, precision=3))
        [ 1201098.489  3299990.978 -5325574.803]
        """
        return self._internal.altitude_intercepts(
            altitude,
            np.atleast_1d(observer).astype(np.float64),
            np.atleast_1d(look_vector).astype(np.float64),
        )


class WGS84(Geodetic):
    def __init__(self):
        """
        A geodetic object based upon the standard WGS84 ellipsoid. See
        :py:class:`sasktran2.Geodetic` for more information and usage.
        """
        super().__init__(6378137.0, 1.0 / 298.257223563)

    def __repr__(self):
        if self.valid:
            return f"WGS84 Location:\nLatitude: {self.latitude}, Longitude: {self.longitude}, Altitude: {self.altitude}"
        else:  # noqa: RET505
            return "WGS84 Unitialized"


class SphericalGeoid(Geodetic):
    def __init__(self, radius: float):
        """
        A geoid that is represented as a perfect sphere. See
        :py:class:`sasktran2.Geodetic` for more information and usage.

        Parameters
        ----------
        radius: float
            Radius of the sphere in [m]
        """
        super().__init__(radius, 0.0)

    def __repr__(self):
        if self.valid:
            return f"Spherical Geoid Location:\nLatitude: {self.latitude}, Longitude: {self.longitude}, Altitude: {self.altitude}"
        else:  # noqa: RET505
            return "Spherical Geoid Unitialized"



from __future__ import annotations

import numpy as np
import xarray as xr

import sasktran2 as sk
from sasktran2._core_rust import PyEngine
from sasktran2.viewinggeo.base import ViewingGeometryContainer


def map_surface_derivative(
    mapping, np_deriv: np.ndarray, dims: list[str]
) -> xr.DataArray:
    if mapping.interpolator is None or len(mapping.interpolator) == 0:
        return xr.DataArray(np_deriv, dims=dims)
    return xr.DataArray(
        np.einsum(
            "ijk, il->lijk",
            np_deriv,
            mapping.interpolator,
            optimize=True,
        ),
        dims=[mapping.interp_dim, *dims],
    )


class Engine:
    _engine: PyEngine

    def __init__(
        self,
        config: sk.Config,
        geometry: sk.Geometry1D,
        viewing_geometry: sk.ViewingGeometry,
    ):
        """
        An Engine is the main class that handles the radiative transfer calculation.  The calculation takes
        place in two components.

        First, upon construction of the Engine, the majority of the geometry information is computed and
        cached.

        The main calculation takes place when calling :py:meth:`~calculate_radiance` with an
        :py:class:`sasktran2.Atmosphere` object where the actual radiative transfer calculation
        is performed.

        Parameters
        ----------
        config : sk.Config
            Configuration object
        model_geometry : sk.Geometry1D
            Geometry for the model
        viewing_geo : sk.ViewingGeometry
            Viewing geometry
        """
        self._engine = PyEngine(
            config._config, geometry._geometry, viewing_geometry._viewing_geometry
        )
        self._config = config
        self._geometry = geometry
        self._viewing_geometry = viewing_geometry

    def calculate_radiance(self, atmosphere: sk.Atmosphere) -> xr.Dataset:
        """
        Performs the radiative transfer calculation for a given atmosphere

        Parameters
        ----------
        atmosphere : sk.Atmosphere
            The atmosphere object containing the atmospheric profile and constituents

        Returns
        -------
        xr.Dataset
            An xarray dataset containing the radiance and derivatives
        """
        output = self._engine.calculate_radiance(atmosphere.internal_object())

        out_ds = xr.Dataset()

        out_ds["radiance"] = xr.DataArray(
            output.radiance,
            dims=["wavelength", "los", "stokes"],
        )

        if atmosphere.wavelengths_nm is not None:
            out_ds.coords["wavelength"] = atmosphere.wavelengths_nm

        out_ds.coords["stokes"] = ["I", "Q", "U", "V"][: len(out_ds.stokes)]

        for k, v in output.d_radiance.items():
            mapping = atmosphere.storage.get_derivative_mapping(k)

            name = k if mapping.assign_name == "" else mapping.assign_name

            if name in out_ds:
                out_ds[name] += v
            else:
                out_ds[name] = xr.DataArray(
                    v,
                    dims=[mapping.interp_dim, "wavelength", "los", "stokes"],
                )
        for k, v in output.d_radiance_surf.items():
            mapping = atmosphere.surface.get_derivative_mapping(k)

            mapped_derivative = map_surface_derivative(
                mapping, v, ["wavelength", "los", "stokes"]
            )
            if mapping.interp_dim == "dummy":
                mapped_derivative = mapped_derivative.isel(**{mapping.interp_dim: 0})
            out_ds[k] = mapped_derivative

        if isinstance(self._viewing_geometry, ViewingGeometryContainer):
            out_ds = self._viewing_geometry.add_geometry_to_radiance(out_ds)

        if self._config.output_los_optical_depth:
            los_od = output.los_optical_depth
            out_ds["los_optical_depth"] = xr.DataArray(
                los_od,
                dims=["wavelength", "los"],
            )

        return out_ds



# ruff: noqa: F401, E402
from __future__ import annotations

import os

os.environ.setdefault("OPENBLAS_NUM_THREADS", "1")
os.environ.setdefault("RUST_BACKTRACE", "full")

# from ._core import Geodetic

from . import (
    appconfig,
    climatology,
    constituent,
    database,
    mie,
    optical,
    solar,
    spectroscopy,
    test_util,
    util,
    viewinggeo,
)
from ._core_rust import (
    EmissionSource,
    GeometryType,
    InputValidationMode,
    InterpolationMethod,
    MultipleScatterSource,
    OccultationSource,
    SingleScatterSource,
    StokesBasis,
    ThreadingLib,
    ThreadingModel,
)
from .atmosphere import Atmosphere
from .config import Config
from .engine import Engine
from .geodetic import WGS84, Geodetic, SphericalGeoid
from .geometry import Geometry1D
from .output import Output, OutputDerivMapped, OutputIdeal
from .viewinggeo.wrappers import (
    GroundViewingSolar,
    SolarAnglesObserverLocation,
    TangentAltitudeSolar,
    ViewingGeometry,
)



from __future__ import annotations

import numpy as np

from sasktran2._core_rust import PyLineAbsorber
from sasktran2.database.aer_line import AERLineDatabase
from sasktran2.optical.base import OpticalProperty
from sasktran2.util import get_hapi


class AERLineAbsorber(OpticalProperty):
    _internal: PyLineAbsorber

    def __init__(self, molecule: str, cull_factor: float = 0.0, line_coupling=False):
        hapi = get_hapi()

        # Ensure the db is downloaded
        db = AERLineDatabase()
        db_path = db.path(None)

        self._internal = PyLineAbsorber(
            molecule,
            db_path.as_posix(),
            cull_factor,
            line_coupling,
            hapi.partitionSum,
            hapi.molecularMass,
        )

    def atmosphere_quantities(self, atmo, **kwargs):
        return self._internal.atmosphere_quantities(atmo)

    def _into_rust_object(self):
        return self._internal

    def cross_sections(self, wavelengths_nm, _altitudes_m, **kwargs):
        pressure_pa = kwargs.get("pressure_pa")
        if pressure_pa is None:
            msg = "Pressure must be provided in the kwargs"
            raise ValueError(msg)

        temperature_k = kwargs.get("temperature_k")
        if temperature_k is None:
            msg = "Temperature must be provided in the kwargs"
            raise ValueError(msg)

        p_self = kwargs.get("p_self")
        if p_self is None:
            p_self = np.zeros_like(pressure_pa)

        return self._internal.cross_section(
            1e7 / wavelengths_nm,
            np.atleast_1d(temperature_k).astype(float),
            np.atleast_1d(pressure_pa).astype(float),
            np.atleast_1d(p_self).astype(float),
        )



from __future__ import annotations

from pathlib import Path

import numpy as np
import xarray as xr

from sasktran2._core_rust import (
    AbsorberDatabaseDim1,
    AbsorberDatabaseDim2,
    AbsorberDatabaseDim3,
)
from sasktran2.atmosphere import Atmosphere, NativeGridDerivative
from sasktran2.optical.base import OpticalProperty, OpticalQuantities
from sasktran2.polarization import LegendreStorageView


class OpticalDatabase(OpticalProperty):
    def __init__(self, db_filepath: Path) -> None:
        """
        An optical property that is defined by a database file.  This is just a base class to handle file loading,
        derived classes must be used as actual optical properties.

        Parameters
        ----------
        db_filepath : Path
            Path to the optical database file

        Raises
        ------
        OSError
            If xarray is not installed
        """
        super().__init__()
        self._file = db_filepath

        try:
            import xarray as xr
        except ImportError:
            msg = "xarray must be installed to use OpticalDatabaseGenericAbsorber"
            raise msg from OSError

        self._database = xr.open_dataset(self._file)
        self._validate_db()


class OpticalDatabaseGenericAbsorber(OpticalDatabase):
    def __init__(self, db_filepath: Path) -> None:
        """
        A purely absorbing optical property defined by a database file.  The database must contain the following

        - xs : The absorption cross section in [m^2]

        xs must be a function of either wavelength_nm or wavenumber_cminv, and optionally temperature and pressure.

        Parameters
        ----------
        db_filepath : Path
            Path to the database file
        """
        super().__init__(db_filepath)

    def _validate_db(self):
        if type(self._database) is not xr.Dataset:
            return
        # Old file format used "temperature" and "pressure" instead of the standard keys
        # "temperature_k" and "pressure_pa", if so we rename them

        if "temperature" in self._database:
            self._database = self._database.rename({"temperature": "temperature_k"})
        if "pressure" in self._database:
            self._database = self._database.rename({"pressure": "pressure_pa"})

        if "xs" not in self._database:
            msg = "xs must be defined in the optical database"
            raise ValueError(msg)

        coords = list(self._database["xs"].coords)
        if "wavenumber_cminv" in coords:
            wavenumber_cminv = self._database["wavenumber_cminv"].to_numpy()
        else:
            wavenumber_cminv = 1e7 / self._database["wavelength_nm"].to_numpy()
            coords[-1] = "wavenumber_cminv"

        if len(coords) == 3:
            param0 = self._database[coords[0]].to_numpy()
            param1 = self._database[coords[1]].to_numpy()

            sidx0 = np.argsort(param0)
            sidx1 = np.argsort(param1)
            sidx2 = np.argsort(wavenumber_cminv)

            xs = self._database["xs"].to_numpy()[sidx0][:, sidx1][:, :, sidx2].copy()

            self._database = AbsorberDatabaseDim3(
                wavenumber_cminv[sidx2].astype(np.float64),
                param0[sidx0].astype(np.float64),
                param1[sidx1].astype(np.float64),
                xs,
                coords[:-1],
            )
            self._coords = coords
        elif len(coords) == 2:
            param0 = self._database[coords[0]].to_numpy()

            sidx0 = np.argsort(param0)
            sidx1 = np.argsort(wavenumber_cminv)

            xs = self._database["xs"].to_numpy()[sidx0][:, sidx1].copy()

            self._database = AbsorberDatabaseDim2(
                wavenumber_cminv[sidx1].astype(np.float64),
                param0[sidx0].astype(np.float64),
                xs,
                coords[:-1],
            )
            self._coords = coords
        elif len(coords) == 1:
            sidx1 = np.argsort(wavenumber_cminv)
            xs = self._database["xs"].to_numpy()[sidx1].copy()

            self._database = AbsorberDatabaseDim1(
                wavenumber_cminv[sidx1].astype(np.float64),
                xs,
            )
            self._coords = coords

    def atmosphere_quantities(self, atmo, **kwargs):
        # When called by the constituent this should be elided
        return self._database.atmosphere_quantities(atmo)

    def optical_derivatives(self, atmo, **kwargs):
        # When called by the constituent this should be elided
        return self._database.optical_derivatives(atmo)

    def _into_rust_object(self):
        return self._database


class OpticalDatabaseGenericScatterer(OpticalDatabase):
    def __init__(self, db_filepath: Path) -> None:
        """
        A purely scattering optical property defined by a database file.  The database must contain the following

        - xs_total : The total cross section in [m^2]
        - xs_scattering : The scattering cross section in [m^2]
        - lm_a1 : the legendre coefficients for the phase function

        All variables must be a function of either wavelength_nm or wavenumber_cminv, and optionally any other dimension such
        as particle size.

        Parameters
        ----------
        db_filepath : Path
            Path to the database file
        """
        super().__init__(db_filepath)

        self._validate_db()

    def _validate_db(self):
        self._database["lm_a1"] /= self._database["lm_a1"].isel(legendre=0)

    def _construct_interp_handler(self, atmo: Atmosphere, **kwargs) -> dict:
        coords = self._database["xs_total"].coords
        interp_handler = {}

        if "wavelength_nm" in coords:
            if atmo.wavelengths_nm is None:
                msg = "wavelengths_nm must be specified in Atmosphere to use OpticalDatabaseGenericScatterer"
                raise ValueError(msg)
            interp_handler["wavelength_nm"] = atmo.wavelengths_nm

        if "wavenumber_cminv" in coords:
            if atmo.wavenumber_cminv is None:
                msg = "wavenumber_cminv must be specified in Atmosphere to use OpticalDatabaseGenericScatterer"
                raise ValueError(msg)
            interp_handler["wavenumber_cminv"] = atmo.wavenumber_cminv

        for name, vals in kwargs.items():
            if name in coords:
                interp_handler[name] = ("z", vals)

        return interp_handler

    def cross_sections(
        self, wavelengths_nm: np.array, altitudes_m: np.array, **kwargs  # noqa: ARG002
    ) -> OpticalQuantities:
        quants = OpticalQuantities()

        coords = self._database["xs_total"].coords
        interp_handler = {}

        interp_handler["wavelength_nm"] = wavelengths_nm

        for name, vals in kwargs.items():
            if name in coords:
                interp_handler[name] = ("z", vals)

        ds_interp = self._database.interp(**interp_handler)

        quants.extinction = ds_interp["xs_total"].to_numpy()
        quants.ssa = ds_interp["xs_scattering"].to_numpy() / quants.extinction

        quants.extinction[np.isnan(quants.extinction)] = 0
        quants.ssa[np.isnan(quants.ssa)] = 0

        return quants

    def atmosphere_quantities(self, atmo: Atmosphere, **kwargs) -> OpticalQuantities:
        quants = OpticalQuantities()

        interp_handler = self._construct_interp_handler(atmo, **kwargs)

        num_assign_legendre = min(
            atmo.leg_coeff.a1.shape[0], len(self._database["legendre"])
        )

        if atmo.nstokes == 1:
            drop_vars = ["lm_a2", "lm_a3", "lm_a4", "lm_b1", "lm_b2"]
        elif atmo.nstokes == 3:
            drop_vars = ["lm_a4", "lm_b2"]
        else:
            drop_vars = []

        ds_interp = (
            self._database.isel(legendre=slice(0, num_assign_legendre))
            .drop_vars(drop_vars)
            .interp(**interp_handler)
        )

        if "z" not in ds_interp.dims:
            # Our dataset has no z dependence
            ds_interp = ds_interp.expand_dims(
                dim={"z": len(atmo.model_geometry.altitudes())}
            )

        quants.extinction = np.copy(
            ds_interp["xs_total"].transpose("z", "wavelength_nm").to_numpy()
        )
        quants.ssa = np.copy(
            ds_interp["xs_scattering"].transpose("z", "wavelength_nm").to_numpy()
        )

        quants.leg_coeff = np.zeros_like(atmo.storage.leg_coeff)

        leg_coeff = LegendreStorageView(quants.leg_coeff, atmo.nstokes)

        leg_coeff.a1[:] = (
            (ds_interp["lm_a1"])
            .transpose("legendre", "z", "wavelength_nm")
            .to_numpy()[:num_assign_legendre]
        )

        # Renormalize leg_coeffs so a1 is always 1
        leg_coeff.a1[:] /= quants.leg_coeff[0, :, :][np.newaxis, :, :]

        quants.extinction[np.isnan(quants.extinction)] = 0
        quants.ssa[np.isnan(quants.ssa)] = 0

        if atmo.nstokes == 3:
            # TODO: add pol properties

            leg_coeff.a2[:] = (
                (ds_interp["lm_a2"])
                .transpose("legendre", "z", "wavelength_nm")
                .to_numpy()[:num_assign_legendre]
            )

            leg_coeff.a3[:] = (
                (ds_interp["lm_a3"])
                .transpose("legendre", "z", "wavelength_nm")
                .to_numpy()[:num_assign_legendre]
            )

            leg_coeff.b1[:] = (
                (ds_interp["lm_b1"])
                .transpose("legendre", "z", "wavelength_nm")
                .to_numpy()[:num_assign_legendre]
            )

        quants.leg_coeff[np.isnan(quants.leg_coeff)] = 0

        return quants

    def optical_derivatives(self, atmo: Atmosphere, **kwargs) -> dict:
        derivs = {}

        interp_handler = self._construct_interp_handler(atmo, **kwargs)

        # Split the interpolators into ones that are 'z' dependent and ones that are not

        interp_handler_z = {}
        interp_handler_noz = {}

        for key, val in interp_handler.items():
            if val[0] == "z":
                interp_handler_z[key] = val
            else:
                interp_handler_noz[key] = val
        num_assign_legendre = min(
            atmo.leg_coeff.a1.shape[0], len(self._database["legendre"])
        )

        if atmo.nstokes == 1:
            drop_vars = ["lm_a2", "lm_a3", "lm_a4", "lm_b1", "lm_b2"]
        elif atmo.nstokes == 3:
            drop_vars = ["lm_a4", "lm_b2"]
        else:
            drop_vars = []

        # Get the derivatives of the cross section with respect to the z dependent variables
        partial_interp = (
            self._database.isel(legendre=slice(0, num_assign_legendre))
            .drop_vars(drop_vars)
            .interp(**interp_handler_noz)
        )

        for key, val in interp_handler_z.items():
            # If the db only contains one element in the dimension we can't take a derivative
            if len(self._database[key]) == 1:
                continue

            # Interpolate over the other variables
            new_interpolants = {k: v for k, v in interp_handler_z.items() if k != key}

            partial_interp2 = partial_interp.interp(**new_interpolants)

            dT = partial_interp2.diff(key) / partial_interp2[key].diff(key)

            interp_index = np.argmax(dT[key].to_numpy() > val[1][:, np.newaxis], axis=1)

            if "z" in dT.dims:
                dT = dT.isel(
                    {
                        "z": xr.DataArray(list(range(len(interp_index))), dims="z"),
                        key: xr.DataArray(interp_index, dims="z"),
                    }
                )
            else:
                dT = dT.isel({key: xr.DataArray(interp_index, dims="z")})

            derivs[key] = NativeGridDerivative(
                d_extinction=dT["xs_total"].transpose("z", "wavelength_nm").to_numpy(),
                d_ssa=dT["xs_scattering"].transpose("z", "wavelength_nm").to_numpy(),
                d_leg_coeff=np.zeros_like(atmo.storage.leg_coeff),
            )

            d_leg_coeff = LegendreStorageView(derivs[key].d_leg_coeff, atmo.nstokes)

            d_leg_coeff.a1[:] = (
                dT["lm_a1"]
                .transpose("legendre", "z", "wavelength_nm")
                .to_numpy()[:num_assign_legendre]
            )

            if atmo.nstokes == 3:
                d_leg_coeff.a2[:] = (
                    dT["lm_a2"]
                    .transpose("legendre", "z", "wavelength_nm")
                    .to_numpy()[:num_assign_legendre]
                )
                d_leg_coeff.a3[:] = (
                    dT["lm_a3"]
                    .transpose("legendre", "z", "wavelength_nm")
                    .to_numpy()[:num_assign_legendre]
                )
                d_leg_coeff.b1[:] = (
                    dT["lm_b1"]
                    .transpose("legendre", "z", "wavelength_nm")
                    .to_numpy()[:num_assign_legendre]
                )

            derivs[key].d_extinction[np.isnan(derivs[key].d_extinction)] = 0
            derivs[key].d_ssa[np.isnan(derivs[key].d_ssa)] = 0
            derivs[key].d_leg_coeff[np.isnan(derivs[key].d_leg_coeff)] = 0

        return derivs

    def cross_section_derivatives(
        self, wavelengths_nm: np.array, altitudes_m: np.array, **kwargs  # noqa: ARG002
    ) -> dict:
        derivs = {}
        coords = self._database["xs_total"].coords
        interp_handler = {}

        interp_handler["wavelength_nm"] = wavelengths_nm

        for name, vals in kwargs.items():
            if name in coords:
                interp_handler[name] = ("z", vals)

        # Split the interpolators into ones that are 'z' dependent and ones that are not

        interp_handler_z = {}
        interp_handler_noz = {}

        for key, val in interp_handler.items():
            if val[0] == "z":
                interp_handler_z[key] = val
            else:
                interp_handler_noz[key] = val

        # Get the derivatives of the cross section with respect to the z dependent variables
        partial_interp = self._database["xs_total"].interp(**interp_handler_noz)

        for key, val in interp_handler_z.items():
            # If the db only contains one element in the dimension we can't take a derivative
            if len(self._database[key]) == 1:
                continue

            # Interpolate over the other variables
            new_interpolants = {k: v for k, v in interp_handler_z.items() if k != key}

            partial_interp2 = partial_interp.interp(**new_interpolants)

            dT = partial_interp2.diff(key) / partial_interp2[key].diff(key)

            interp_index = np.argmax(dT[key].to_numpy() > val[1][:, np.newaxis], axis=1)

            if "z" in dT.dims:
                dT = dT.isel(
                    {
                        "z": xr.DataArray(list(range(len(interp_index))), dims="z"),
                        key: xr.DataArray(interp_index, dims="z"),
                    }
                )
            else:
                dT = dT.isel({key: xr.DataArray(interp_index, dims="z")})

            derivs[key] = dT.to_numpy().flatten()

        return derivs



from __future__ import annotations

import hashlib
import json
import logging
from pathlib import Path

import numpy as np
import xarray as xr

from sasktran2.optical.database import (
    OpticalDatabaseGenericAbsorber,
)
from sasktran2.optical.hitran import HITRANAbsorber
from sasktran2.units import wavenumber_cminv_to_wavlength_nm
from sasktran2.util import get_hapi

from .base import CachedDatabase
from .hitran_line import HITRANLineDatabase

PRESSURE_GRID = (
    np.array(
        [
            1.03181655e03,
            9.99910226e02,
            9.72231997e02,
            9.44960640e02,
            9.18098202e02,
            8.91646636e02,
            8.65607893e02,
            8.39983721e02,
            8.14775620e02,
            7.89985094e02,
            7.65613488e02,
            7.41661804e02,
            7.18131094e02,
            6.95022257e02,
            6.72335891e02,
            6.50072549e02,
            6.28232578e02,
            6.06816233e02,
            5.85823558e02,
            5.65254408e02,
            5.45108580e02,
            5.25385676e02,
            5.06085044e02,
            4.87205987e02,
            4.68747652e02,
            4.50708890e02,
            4.33088502e02,
            4.15885138e02,
            3.99097246e02,
            3.82723078e02,
            3.66760783e02,
            3.51208311e02,
            3.36063513e02,
            3.21323986e02,
            3.06987184e02,
            2.93050454e02,
            2.79510897e02,
            2.66365512e02,
            2.53611100e02,
            2.41244361e02,
            2.29261844e02,
            2.17659797e02,
            2.06434425e02,
            1.95581772e02,
            1.85097641e02,
            1.74977731e02,
            1.65217644e02,
            1.55812775e02,
            1.46758379e02,
            1.38049500e02,
            1.29681042e02,
            1.21647854e02,
            1.13944585e02,
            1.06565734e02,
            9.95056250e01,
            9.27584798e01,
            8.63183864e01,
            8.01792901e01,
            7.43350047e01,
            6.87792048e01,
            6.35054499e01,
            5.85071744e01,
            5.37776926e01,
            4.93102089e01,
            4.50978073e01,
            4.11334673e01,
            3.74100678e01,
            3.39203935e01,
            3.06571276e01,
            2.76128649e01,
            2.47801240e01,
            2.21513493e01,
            1.97189141e01,
            1.74751173e01,
            1.54122078e01,
            1.35223946e01,
            1.17978408e01,
            1.02306666e01,
            8.81297618e00,
            7.53686238e00,
            6.39440891e00,
            5.37770790e00,
            4.47887025e00,
            3.69003722e00,
            3.00339331e00,
            2.41118145e00,
            1.90571877e00,
            1.47941371e00,
            1.12478403e00,
            8.34477191e-01,
            6.01292929e-01,
            4.18208541e-01,
            2.78406802e-01,
            1.75308171e-01,
            1.02608084e-01,
            5.43204621e-02,
        ]
    )
    * 100
)
TEMP_GRID = np.arange(190, 311, 10)


class HITRANDatabase(CachedDatabase, OpticalDatabaseGenericAbsorber):
    def __init__(
        self,
        molecule: str,
        start_wavenumber: float,
        end_wavenumber: float,
        wavenumber_resolution: float,
        reduction_factor: int,
        num_threads: int = 1,
        db_root: Path | None = None,
        backend: str = "sasktran2",
        profile: str = "voigt",
    ) -> None:
        """


        Parameters
        ----------
        molecule : str
            The HITRAN molecule identifier, e.g. "CO2", "H2O", "O3", "CH4"
        start_wavenumber : float
            The start wavenumber of the database in cm^-1
        end_wavenumber : float
            The end wavenumber of the database in cm^-1
        wavenumber_resolution : float
            The wavenumber resolution of the database in cm^-1
        num_threads: int, optional
            The number of threads to use in the cross section calculation. Default is 1.
        db_root : Path, optional
            The root directory to store the database, by default None
        backend : str, optional
            The backend to use, by default "sasktran2" which uses the internal broadening algorithms.
            options are ["sasktran2", "sasktran_legacy", "hapi"]. Default is "sasktran2".
            Note: "sasktran_legacy" requires the sasktran package to be installed.
            Both the "sasktran2" and "hapi" backends use the hapi package to access the HITRAN database.
        profile : str, optional
            The line shape profile to use in the cross section calculation. Currently, only the "hapi" backend supports options other
            than the default "voigt". Supported options are ["voigt", "sdvoigt", "ht", "priority", "lorentz", "doppler"].
            Default is "voigt".
        """

        class NumpyEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, np.ndarray):
                    return obj.tolist()
                return json.JSONEncoder.default(self, obj)

        self._profile = profile.lower()
        if backend == "sasktran_legacy" and self._profile != "voigt":
            logging.warning(
                "%s line profile is not supported by sasktran_legacy backend. Switching to voigt",
                self._profile,
            )
            self._profile = "voigt"

        hasher = hashlib.sha1()
        encoded = json.dumps(
            {
                "start_wavenumber": start_wavenumber,
                "end_wavenumber": end_wavenumber,
                "wavenumber_resolution": wavenumber_resolution,
                "reduction_factor": reduction_factor,
                "profile": self._profile,
            },
            sort_keys=True,
            cls=NumpyEncoder,
        ).encode()
        hasher.update(encoded)
        identifier = hasher.hexdigest()
        self._backend = backend
        self._start_wavenumber = start_wavenumber
        self._end_wavenumber = end_wavenumber
        self._wavenumber_resolution = wavenumber_resolution
        self._reduction_factor = reduction_factor
        self._molecule = molecule
        self._ltol = 1e-9
        self._wavenumber_wing = 50
        self._wavenumber_margin = 10
        self._num_threads = num_threads

        CachedDatabase.__init__(
            self,
            db_root=db_root,
            rel_path=Path("hitran").joinpath(self._molecule, self._backend),
        )

        self._data_file = self._db_root.joinpath(f"{identifier}.nc")

        OpticalDatabaseGenericAbsorber.__init__(self, self.path())

    def generate(self):
        if self._backend == "sasktran2":
            self._generate_sasktran2()
        elif self._backend == "sasktran_legacy":
            self._generate_sasktran_legacy()
        elif self._backend == "hapi":
            self._generate_hapi()
        else:
            msg = f"Invalid backend {self._backend}"
            raise ValueError(msg)

    def clear(self):
        if self._data_file.exists():
            self._data_file.unlink()

    def path(self, key: str | None = None, **kwargs) -> Path | None:  # noqa: ARG002
        if not self._data_file.exists():
            self.generate()
        return self._data_file

    def load_ds(self, key: str | None = None, **kwargs) -> xr.Dataset:
        return xr.open_dataset(self.path(key, **kwargs))

    def _generate_sasktran_legacy(self):
        try:
            from sasktran import ClimatologyUserDefined, HITRANChemical
        except ImportError as err:
            msg = "sasktran_legacy is required to generate Mie databases, try pip install sasktran"
            raise ImportError(msg) from err

        hires_wavenumber_grid = np.arange(
            self._start_wavenumber, self._end_wavenumber, self._wavenumber_resolution
        )

        xs = np.zeros((len(PRESSURE_GRID), len(TEMP_GRID), len(hires_wavenumber_grid)))

        for idx, pres in enumerate(PRESSURE_GRID):
            for idy, temp in enumerate(TEMP_GRID):
                xs[idx, idy] = (
                    HITRANChemical(self._molecule, line_tolerance=self._ltol)
                    .calculate_cross_sections(
                        ClimatologyUserDefined(
                            altitudes=[0, 100000],
                            values={
                                "SKCLIMATOLOGY_PRESSURE_PA": [pres, pres],
                                "SKCLIMATOLOGY_TEMPERATURE_K": [temp, temp],
                            },
                        ),
                        0,
                        0,
                        10000,
                        54372,
                        wavenumber_cminv_to_wavlength_nm(hires_wavenumber_grid),
                    )
                    .total
                    / 1e4
                )

        ds = xr.Dataset(
            {"xs": (["pressure", "temperature", "wavenumber_cminv"], xs)},
            coords={
                "pressure": PRESSURE_GRID,
                "temperature": TEMP_GRID,
                "wavenumber_cminv": hires_wavenumber_grid,
            },
        )

        ds.to_netcdf(self._data_file)

    def _generate_hapi(self):
        line_db = HITRANLineDatabase()
        line_db.path(self._molecule)

        hapi = get_hapi()

        abs_cal_fn = {
            "voigt": hapi.absorptionCoefficient_Voigt,
            "sdvoigt": hapi.absorptionCoefficient_SDVoigt,
            "ht": hapi.absorptionCoefficient_HT,
            "priority": hapi.absorptionCoefficient_Priority,
            "lorentz": hapi.absorptionCoefficient_Lorentz,
            "doppler": hapi.absorptionCoefficient_Doppler,
        }
        if self._profile not in abs_cal_fn:
            msg = f"Invalid line profile {self._profile}"
            raise ValueError(msg)

        line_db.initialize_hapi(self._molecule)

        hapi.select(
            self._molecule,
            DestinationTableName="spectral_window",
            Conditions=(
                "between",
                "nu",
                self._start_wavenumber - self._wavenumber_margin,
                self._end_wavenumber + self._wavenumber_margin,
            ),
        )

        hires_wavenumber_grid = np.arange(
            self._start_wavenumber, self._end_wavenumber, self._wavenumber_resolution
        )

        xs = np.zeros((len(PRESSURE_GRID), len(TEMP_GRID), len(hires_wavenumber_grid)))

        for idx, pres in enumerate(PRESSURE_GRID):
            for idy, temp in enumerate(TEMP_GRID):
                _, xs_hapi = abs_cal_fn[self._profile](
                    SourceTables="spectral_window",
                    Environment={"T": temp, "p": pres / 101325.0},
                    WavenumberGrid=hires_wavenumber_grid.tolist(),
                    WavenumberWing=self._wavenumber_wing,
                )
                xs[idx, idy] = xs_hapi / 1e4

        ds = xr.Dataset(
            {"xs": (["pressure", "temperature", "wavenumber_cminv"], xs)},
            coords={
                "pressure": PRESSURE_GRID,
                "temperature": TEMP_GRID,
                "wavenumber_cminv": hires_wavenumber_grid,
            },
        )

        ds.to_netcdf(self._data_file)

    def _generate_sasktran2(self):
        hires_wavenumber_grid = np.arange(
            self._start_wavenumber, self._end_wavenumber, self._wavenumber_resolution
        )

        calculator = HITRANAbsorber(
            self._molecule, line_contribution_width=self._wavenumber_wing / 2
        )
        xs = np.zeros((len(PRESSURE_GRID), len(TEMP_GRID), len(hires_wavenumber_grid)))

        for i, temperature in enumerate(TEMP_GRID):
            xs[:, i, :] = calculator.cross_sections(
                temperature_k=np.ones(len(PRESSURE_GRID)) * temperature,
                pressure_pa=PRESSURE_GRID,
                wavelengths_nm=wavenumber_cminv_to_wavlength_nm(hires_wavenumber_grid),
                num_threads=self._num_threads,
                altitudes_m=np.ones(len(PRESSURE_GRID)) * 0,
            ).T
        ds = xr.Dataset(
            {"xs": (["pressure", "temperature", "wavenumber_cminv"], xs)},
            coords={
                "pressure": PRESSURE_GRID,
                "temperature": TEMP_GRID,
                "wavenumber_cminv": hires_wavenumber_grid,
            },
        )

        ds.to_netcdf(self._data_file)
